"""Tests that metsrw can create normative logical structMaps correctly.

A normative logical structMap is a <mets:structMap> with TYPE="logical" and
LABEL="Normative Directory Structure", which documents empty directories that
are absent in the standard physical structural map.

"""

import uuid
from unittest import TestCase

from lxml import etree

import metsrw
from metsrw.plugins.premisrw import PREMIS_3_0_NAMESPACES
from metsrw.plugins.premisrw import PREMISObject
from metsrw.plugins.premisrw import lxmlns


class TestNormativeStructMap(TestCase):
    """Test normative logical structmap class."""

    def test_read_normative_structmaps(self):
        """It should be able to read a normative logical structural map.

        When parsing a mets file that contains a logical structmap with label
        "Normative Directory Structure", metsrw should find the empty
        directories in that logical structmap and create ``FSEntry`` instances
        for them. During (re-)serialization, the physical structmap should not
        document the empty directory but the normative (logical) one should.
        """
        mw = metsrw.METSDocument.fromfile("fixtures/mets_empty_dirs.xml")
        empty_dir_filesec = mw.get_file(label="dir2aiii", type="Directory")
        assert empty_dir_filesec is not None

        structmap = mw._structmap()
        normative_structmap = mw._normative_structmap()
        exists_only_in_normative = (
            'mets:div[@LABEL="thu1-31d2689d-91d0-42b4-91be-1b1b0c2e65a4"]/'
            'mets:div[@LABEL="objects"]/'
            'mets:div[@LABEL="dir2"]/'
            'mets:div[@LABEL="dir2a"]/'
            'mets:div[@LABEL="dir2aiii"]'
        )
        exists_in_both = (
            'mets:div[@LABEL="thu1-31d2689d-91d0-42b4-91be-1b1b0c2e65a4"]/'
            'mets:div[@LABEL="objects"]/'
            'mets:div[@LABEL="dir2"]/'
            'mets:div[@LABEL="dir2a"]/'
            'mets:div[@LABEL="dir2aii"]'
        )
        assert structmap.find(exists_only_in_normative, metsrw.NAMESPACES) is None
        assert structmap.find(exists_in_both, metsrw.NAMESPACES) is not None
        assert (
            normative_structmap.find(exists_only_in_normative, metsrw.NAMESPACES)
            is not None
        )
        assert normative_structmap.find(exists_in_both, metsrw.NAMESPACES) is not None

    def test_write_normative_structmap(self):
        """It should be able to write a normative logical structural map."""

        # Create the empty directory as an FSEntry and give it a simple PREMIS
        # object.
        d_empty = metsrw.FSEntry("EMPTY_DIR", type="Directory")
        d_empty_id = str(uuid.uuid4())
        d_empty_premis_object = PREMISObject(
            identifier_value=d_empty_id,
            premis_version="3.0",
            xsi_type="premis:intellectualEntity",
        )
        d_empty.add_premis_object(d_empty_premis_object)

        # Create the parent directory of the empty directory and give it a
        # simple PREMIS object also.
        f3 = metsrw.FSEntry("level3.txt", file_uuid=str(uuid.uuid4()))
        f3.add_dmdsec(etree.Element("data"), "OTHER")
        d2 = metsrw.FSEntry("dir2", type="Directory", children=[f3, d_empty])
        d2_id = str(uuid.uuid4())
        d2_premis_object = PREMISObject(identifier_value=d2_id)
        d2.add_premis_object(d2_premis_object)

        # Create more directories and files and add the root of the dir
        # structure to a metsrw METSDocument instance.
        f2 = metsrw.FSEntry("level2.txt", file_uuid=str(uuid.uuid4()))
        d1 = metsrw.FSEntry("dir1", type="Directory", children=[d2, f2])
        f1 = metsrw.FSEntry("level1.txt", file_uuid=str(uuid.uuid4()))
        d = metsrw.FSEntry("root", type="Directory", children=[d1, f1])
        mw = metsrw.METSDocument()
        mw.append_file(d)

        # Expect to find all of our files and directories in the return value
        # of ``all_files``, including the empty directory.
        files = mw.all_files()
        assert files
        assert len(files) == 7
        assert d in files
        assert f1 in files
        assert d1 in files
        assert f2 in files
        assert d2 in files
        assert f3 in files
        assert d_empty in files

        # Get XPaths for the empty directory and its sister file. The file
        # should exist in both structmaps, the empty directory should only
        # exist in the normative logical structmap.
        exists_in_both_path = (
            'mets:div[@LABEL="root"]/'
            'mets:div[@LABEL="dir1"]/'
            'mets:div[@LABEL="dir2"]/'
            'mets:div[@LABEL="level3.txt"]'
        )
        exists_in_normative_only_path = (
            'mets:div[@LABEL="root"]/'
            'mets:div[@LABEL="dir1"]/'
            'mets:div[@LABEL="dir2"]/'
            'mets:div[@LABEL="EMPTY_DIR"]'
        )

        # Expect that the empty directory is not in the physical structmap.
        structmap_el = mw._structmap()
        assert structmap_el.find(exists_in_both_path, metsrw.NAMESPACES) is not None
        assert (
            structmap_el.find(exists_in_normative_only_path, metsrw.NAMESPACES) is None
        )

        # Expect that the empty directory is in the normative logical structmap.
        normative_structmap_el = mw._normative_structmap()
        assert (
            normative_structmap_el.find(exists_in_both_path, metsrw.NAMESPACES)
            is not None
        )
        empty_div_el = normative_structmap_el.find(
            exists_in_normative_only_path, metsrw.NAMESPACES
        )
        assert empty_div_el is not None

        # Expect that the empty directory in the normative logical structmap
        # has a DMDID that references a dmdSec in the METS document and that
        # this dmdSec contains a PREMIS object.
        dmdid = empty_div_el.get("DMDID")
        assert dmdid.startswith("dmdSec_")
        doc = mw.serialize()
        empty_dir_dmd_sec = doc.find(f'mets:dmdSec[@ID="{dmdid}"]', metsrw.NAMESPACES)
        assert empty_dir_dmd_sec is not None
        xml_data_el = empty_dir_dmd_sec.find(
            "mets:mdWrap/mets:xmlData", metsrw.NAMESPACES
        )
        premis_object_el_retrieved = xml_data_el.find(
            "premis:object", PREMIS_3_0_NAMESPACES
        )
        d_empty_id_retrieved = premis_object_el_retrieved.find(
            "premis:objectIdentifier/premis:objectIdentifierValue",
            PREMIS_3_0_NAMESPACES,
        ).text
        assert d_empty_id_retrieved == d_empty_id
        xsi_type = premis_object_el_retrieved.get(
            "{}type".format(lxmlns("xsi", premis_version="3.0"))
        )
        assert xsi_type == "premis:intellectualEntity"

        # Expect that the file in the normative logical structmap includes the
        # DMDID attribute.
        dmdid = normative_structmap_el.find(exists_in_both_path, metsrw.NAMESPACES).get(
            "DMDID"
        )
        assert dmdid.startswith("dmdSec_")
        file_3_dmd_sec = doc.find(f'mets:dmdSec[@ID="{dmdid}"]', metsrw.NAMESPACES)
        assert file_3_dmd_sec is not None



import pytest

import metsrw

GOOD_PATHS_SLASH_URLS = (
    "30_CFLQ_271_13-3-13_1524[1].pdf",
    "30/CFLQ_271_13-3-13_1524[1].pdf",
    r"30\ CFLQ_271_13-3-13_1524[1].pdf",
    "/foo/bar[baz/hello",  # urllib.urlparse will accept this because it's a URL with just a path.
    "http://foobar.com/hello[1].pdf",
)

# urllib.urlparse will choke on these and raise ValueError because of the
# unbalanced bracket in the netloc part.
BAD_URLS = ("http://foo[bar.com/hello[1].pdf",)


def test_url_encoding():
    for url in GOOD_PATHS_SLASH_URLS:
        assert url == metsrw.urldecode(metsrw.urlencode(url))
    for url in BAD_URLS:
        with pytest.raises(ValueError):
            metsrw.urlencode(url)


def test_generate_mdtype_key():
    assert metsrw.generate_mdtype_key("MDTYPE") == "MDTYPE"
    assert metsrw.generate_mdtype_key("MDTYPE", "OTHERMDTYPE") == "MDTYPE_OTHERMDTYPE"



"""PREMIS reader and writer."""

import logging

from .premis import PREMISAgent
from .premis import PREMISElement
from .premis import PREMISEvent
from .premis import PREMISObject
from .premis import PREMISRights
from .premis import data_find
from .premis import data_find_all
from .premis import data_find_text
from .premis import data_find_text_or_all
from .premis import data_to_premis
from .premis import premis_to_data
from .utils import NAMESPACES
from .utils import PREMIS_2_1_META
from .utils import PREMIS_2_1_NAMESPACE
from .utils import PREMIS_2_1_NAMESPACES
from .utils import PREMIS_2_1_SCHEMA_LOCATION
from .utils import PREMIS_2_1_VERSION
from .utils import PREMIS_2_1_XSD
from .utils import PREMIS_2_2_META
from .utils import PREMIS_2_2_NAMESPACE
from .utils import PREMIS_2_2_NAMESPACES
from .utils import PREMIS_2_2_SCHEMA_LOCATION
from .utils import PREMIS_2_2_VERSION
from .utils import PREMIS_2_2_XSD
from .utils import PREMIS_3_0_META
from .utils import PREMIS_3_0_NAMESPACE
from .utils import PREMIS_3_0_NAMESPACES
from .utils import PREMIS_3_0_SCHEMA_LOCATION
from .utils import PREMIS_3_0_VERSION
from .utils import PREMIS_3_0_XSD
from .utils import PREMIS_META
from .utils import PREMIS_SCHEMA_LOCATION
from .utils import PREMIS_VERSION
from .utils import PREMIS_VERSIONS_MAP
from .utils import XSI_NAMESPACE
from .utils import camel_to_snake
from .utils import lxmlns
from .utils import snake_to_camel
from .utils import snake_to_camel_cap

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

__all__ = [
    "PREMISElement",
    "PREMISObject",
    "PREMISEvent",
    "PREMISAgent",
    "PREMISRights",
    "data_to_premis",
    "premis_to_data",
    "data_find",
    "data_find_all",
    "data_find_text",
    "data_find_text_or_all",
    "XSI_NAMESPACE",
    "PREMIS_2_1_VERSION",
    "PREMIS_2_1_NAMESPACE",
    "PREMIS_2_1_XSD",
    "PREMIS_2_1_SCHEMA_LOCATION",
    "PREMIS_2_1_NAMESPACES",
    "PREMIS_2_1_META",
    "PREMIS_2_2_VERSION",
    "PREMIS_2_2_NAMESPACE",
    "PREMIS_2_2_XSD",
    "PREMIS_2_2_SCHEMA_LOCATION",
    "PREMIS_2_2_NAMESPACES",
    "PREMIS_2_2_META",
    "PREMIS_3_0_VERSION",
    "PREMIS_3_0_NAMESPACE",
    "PREMIS_3_0_XSD",
    "PREMIS_3_0_SCHEMA_LOCATION",
    "PREMIS_3_0_NAMESPACES",
    "PREMIS_3_0_META",
    "PREMIS_VERSIONS_MAP",
    "PREMIS_VERSION",
    "NAMESPACES",
    "PREMIS_META",
    "PREMIS_SCHEMA_LOCATION",
    "lxmlns",
    "snake_to_camel_cap",
    "snake_to_camel",
    "camel_to_snake",
]



from unittest import TestCase

import pytest
from lxml import etree

import metsrw


class TestAgent(TestCase):
    def test_parse_exception_on_wrong_tag(self):
        element = etree.Element("test")
        with pytest.raises(metsrw.ParseError):
            metsrw.Agent.parse(element)

    def test_parse_exception_on_missing_role(self):
        element = etree.Element(metsrw.Agent.AGENT_TAG)
        with pytest.raises(metsrw.ParseError):
            metsrw.Agent.parse(element)

    def test_parse_with_role_set(self):
        element = etree.Element(metsrw.Agent.AGENT_TAG, ROLE="CREATOR")
        agent = metsrw.Agent.parse(element)

        assert agent.role == element.get("ROLE")

    def test_parse_with_other_role_set(self):
        element = etree.Element(
            metsrw.Agent.AGENT_TAG, ROLE="OTHER", OTHERROLE="DEVELOPER"
        )
        agent = metsrw.Agent.parse(element)

        assert agent.role == element.get("OTHERROLE")

    def test_parse_with_type_set(self):
        element = etree.Element(
            metsrw.Agent.AGENT_TAG, ROLE="CREATOR", TYPE="INDIVIDUAL"
        )
        agent = metsrw.Agent.parse(element)

        assert agent.type == element.get("TYPE")

    def test_parse_with_other_type_set(self):
        element = etree.Element(
            metsrw.Agent.AGENT_TAG, ROLE="CREATOR", TYPE="OTHER", OTHERTYPE="SOFTWARE"
        )
        agent = metsrw.Agent.parse(element)

        assert agent.type == element.get("OTHERTYPE")

    def test_serialize(self):
        agent = metsrw.Agent(
            "CREATOR",
            id="1",
            type="INDIVIDUAL",
            name="An agent",
            notes=["A Note", "Another Note"],
        )
        element = agent.serialize()

        assert element.get("ID") == agent.id
        assert element.get("ROLE") == agent.role
        assert element.get("TYPE") == agent.type

        assert element[0].text == agent.name
        assert element[1].text == agent.notes[0]
        assert element[2].text == agent.notes[1]

    def test_serialize_with_other_role(self):
        agent = metsrw.Agent("PROGRAMMER")
        element = agent.serialize()

        assert element.get("OTHERROLE") == agent.role


class TestAltRecordId(TestCase):
    def test_parse_exception_on_wrong_tag(self):
        element = etree.Element("test")
        with pytest.raises(metsrw.ParseError):
            metsrw.AltRecordID.parse(element)

    def test_parse(self):
        element = etree.Element(
            metsrw.AltRecordID.ALT_RECORD_ID_TAG, ID="543", TYPE="Test"
        )
        element.text = "a-unique-id"
        alt_record_id = metsrw.AltRecordID.parse(element)

        assert alt_record_id.text == element.text
        assert alt_record_id.id == element.get("ID")
        assert alt_record_id.type == element.get("TYPE")

    def test_serialize(self):
        alt_record_id = metsrw.AltRecordID("12345", id="1", type="Accession Id")
        element = alt_record_id.serialize()

        assert element.get("ID") == alt_record_id.id
        assert element.get("TYPE") == alt_record_id.type

        assert element.text == alt_record_id.text


class TestAMDSec(TestCase):
    """Test AMDSec class."""

    def setUp(self):
        # Reset the id generation counter per test
        metsrw.metadata.AMDSec._id_generator.clear()

    def test_identifier(self):
        amdsec_ids = [metsrw.AMDSec().id_string for _ in range(10)]
        # Generate a SubSection in between to make sure our count
        # doesn't jump
        amdsec_ids.append(metsrw.AMDSec().id_string)

        for index, amdsec_id in enumerate(amdsec_ids, 1):
            assert amdsec_id == f"amdSec_{index}"

    def test_tree_no_id(self):
        with pytest.raises(ValueError) as excinfo:
            metsrw.AMDSec(tree=etree.Element("amdSec"))
        assert "section_id" in str(excinfo.value)

    def test_tree_overwrites_serialize(self):
        elem = etree.Element("temp")
        amdsec = metsrw.AMDSec(tree=elem, section_id="id1")
        assert amdsec.serialize() == elem

    def test_generate_id_skips_existing_id(self):
        element = etree.Element("amdSec")
        amdsec1 = metsrw.AMDSec(tree=element, section_id="amdSec_1")
        amdsec2 = metsrw.AMDSec()
        amdsec3 = metsrw.AMDSec()

        assert amdsec1.id_string != amdsec2.id_string
        assert amdsec1.id_string != amdsec3.id_string
        assert amdsec2.id_string != amdsec3.id_string

        assert metsrw.AMDSec.get_current_id_count() == 3


class TestSubSection(TestCase):
    """Test SubSection class."""

    STUB_MDWRAP = metsrw.MDWrap("<foo/>", "PREMIS:DUMMY")

    def setUp(self):
        # Reset the id generation counters per test
        for counter in metsrw.metadata.SubSection._id_generators.values():
            counter.clear()

    def test_identifier(self):
        tech_md_ids = [metsrw.SubSection("techMD", []).id_string for _ in range(10)]
        dmdsec_ids = [metsrw.SubSection("dmdSec", []).id_string for _ in range(10)]

        for index, tech_md_id in enumerate(tech_md_ids, 1):
            assert tech_md_id == f"techMD_{index}"

        for index, dmdsec_id in enumerate(dmdsec_ids, 1):
            assert dmdsec_id == f"dmdSec_{index}"

    def test_allowed_tags(self):
        """It should only allow certain subsection tags."""
        with pytest.raises(ValueError):
            metsrw.SubSection("fakeMD", None)
        metsrw.SubSection("dmdSec", None)
        metsrw.SubSection("techMD", None)
        metsrw.SubSection("rightsMD", None)
        metsrw.SubSection("sourceMD", None)
        metsrw.SubSection("digiprovMD", None)

    def test_replace_with_diff_type(self):
        """It should assert if replacing with a different subsection type."""
        dmdsec = metsrw.SubSection("dmdSec", self.STUB_MDWRAP)
        rightsmd = metsrw.SubSection("rightsMD", self.STUB_MDWRAP)
        with pytest.raises(metsrw.MetsError):
            dmdsec.replace_with(rightsmd)

    def test_replacement_techmd(self):
        """It should have no special behaviour replacing techMDs."""
        techmd_old = metsrw.SubSection("techMD", self.STUB_MDWRAP)
        techmd_new = metsrw.SubSection("techMD", self.STUB_MDWRAP)
        techmd_old.replace_with(techmd_new)
        assert techmd_old.get_status() == "superseded"
        assert techmd_new.get_status() == "current"

    def test_replacement_sourcemd(self):
        """It should have no special behaviour replacing sourceMDs."""
        sourcemd_old = metsrw.SubSection("sourceMD", self.STUB_MDWRAP)
        sourcemd_new = metsrw.SubSection("sourceMD", self.STUB_MDWRAP)
        sourcemd_old.replace_with(sourcemd_new)
        assert sourcemd_old.get_status() is None
        assert sourcemd_new.get_status() is None

    def test_replacement_digiprovmd(self):
        """It should have no special behaviour replacing digiprovMDs."""
        digiprovmd_old = metsrw.SubSection("digiprovMD", self.STUB_MDWRAP)
        digiprovmd_new = metsrw.SubSection("digiprovMD", self.STUB_MDWRAP)
        digiprovmd_old.replace_with(digiprovmd_new)
        assert digiprovmd_old.get_status() is None
        assert digiprovmd_new.get_status() is None

    def test_replacement_rightsmd(self):
        """
        It should mark the most recent rightsMD 'current'.
        It should mark all other rightsMDs 'superseded'.
        """
        rightsmd_old = metsrw.SubSection("rightsMD", self.STUB_MDWRAP)
        assert rightsmd_old.get_status() == "current"
        rightsmd_new = metsrw.SubSection("rightsMD", self.STUB_MDWRAP)
        rightsmd_old.replace_with(rightsmd_new)
        assert rightsmd_old.get_status() == "superseded"
        assert rightsmd_new.get_status() == "current"
        rightsmd_newer = metsrw.SubSection("rightsMD", self.STUB_MDWRAP)
        rightsmd_new.replace_with(rightsmd_newer)
        assert rightsmd_old.get_status() == "superseded"
        assert rightsmd_new.get_status() == "superseded"
        assert rightsmd_newer.get_status() == "current"

    def test_replacement_dmdsec(self):
        """
        It should mark the first dmdSec 'original'.
        It should mark all other dmdSecs 'updated'.
        """
        dmdsec_old = metsrw.SubSection("dmdSec", self.STUB_MDWRAP)
        assert dmdsec_old.get_status() == "original"
        dmdsec_new = metsrw.SubSection("dmdSec", self.STUB_MDWRAP)
        dmdsec_old.replace_with(dmdsec_new)
        assert dmdsec_old.get_status() == "original-superseded"
        assert dmdsec_new.get_status() == "update"
        dmdsec_newer = metsrw.SubSection("dmdSec", self.STUB_MDWRAP)
        dmdsec_new.replace_with(dmdsec_newer)
        assert dmdsec_old.get_status() == "original-superseded"
        assert dmdsec_new.get_status() == "update-superseded"
        assert dmdsec_newer.get_status() == "update"

    def test_subsection_serialize(self):
        content = metsrw.MDWrap("<foo/>", None)
        content.serialize = lambda: etree.Element("dummy_data")
        subsection = metsrw.SubSection("techMD", content)
        subsection._id = "techMD_1"

        root = subsection.serialize("2014-07-23T21:48:33")
        assert root.tag == "{http://www.loc.gov/METS/}techMD"
        assert root.attrib["ID"] == "techMD_1"
        assert root.attrib["CREATED"] == "2014-07-23T21:48:33"
        assert root.attrib["STATUS"] == "current"
        assert len(root.attrib) == 3
        assert len(root) == 1
        assert root[0].tag == "dummy_data"

    def test_subsection_serialize_no_date(self):
        content = metsrw.MDWrap("<foo/>", None)
        content.serialize = lambda: etree.Element("dummy_data")
        subsection = metsrw.SubSection("techMD", content)
        subsection._id = "techMD_1"

        root = subsection.serialize()
        assert root.tag == "{http://www.loc.gov/METS/}techMD"
        assert root.attrib["ID"] == "techMD_1"
        assert root.attrib["STATUS"] == "current"
        assert len(root.attrib) == 2
        assert len(root) == 1
        assert root[0].tag == "dummy_data"

    def test_subsection_ordering(self):
        list_ = []
        list_.append(metsrw.SubSection("digiprovMD", self.STUB_MDWRAP))
        list_.append(metsrw.SubSection("sourceMD", self.STUB_MDWRAP))
        list_.append(metsrw.SubSection("rightsMD", self.STUB_MDWRAP))
        list_.append(metsrw.SubSection("techMD", self.STUB_MDWRAP))
        list_.sort()
        assert list_[0].subsection == "techMD"
        assert list_[1].subsection == "rightsMD"
        assert list_[2].subsection == "sourceMD"
        assert list_[3].subsection == "digiprovMD"

    def test_parse(self):
        """
        It should parse SubSection ID, CREATED & STATUS.
        It should create an MDWrap or MDRef child.
        """
        # Parses correctly
        elem = etree.Element(
            "{http://www.loc.gov/METS/}techMD",
            ID="techMD_42",
            CREATED="2016-01-02T03:04:05",
            STATUS="original",
        )
        mdr = etree.SubElement(
            elem, "{http://www.loc.gov/METS/}mdRef", MDTYPE="dummy", LOCTYPE="URL"
        )
        mdr.set("{http://www.w3.org/1999/xlink}href", "url")
        obj = metsrw.SubSection.parse(elem)
        assert obj.subsection == "techMD"
        assert obj.status == "original"
        assert obj.created == "2016-01-02T03:04:05"
        assert obj.id_string == "techMD_42"
        assert obj.newer is obj.older is None
        assert isinstance(obj.contents, metsrw.MDRef)

    def test_parse_bad_subsection_type(self):
        """It should only accept valid subsection tags."""
        # Wrong tag name
        bad = etree.Element("foo")
        with pytest.raises(metsrw.ParseError) as e:
            metsrw.SubSection.parse(bad)
            assert "METS namespace" in e.value

    def test_parse_bad_child_type(self):
        """It should only accept valid child tags."""
        elem = etree.Element(
            "{http://www.loc.gov/METS/}techMD",
            ID="techMD_42",
            CREATED="2016-01-02T03:04:05",
            STATUS="original",
        )
        etree.SubElement(elem, "foo")
        with pytest.raises(metsrw.ParseError) as e:
            metsrw.SubSection.parse(elem)
            assert "must be mdWrap or mdRef" in e.value

    def test_roundtrip(self):
        """It should be able to parse and write out a subsection unchanged."""
        elem = etree.Element(
            "{http://www.loc.gov/METS/}techMD",
            ID="techMD_42",
            CREATED="2016-01-02T03:04:05",
            STATUS="original",
        )
        mdr = etree.SubElement(
            elem, "{http://www.loc.gov/METS/}mdRef", MDTYPE="dummy", LOCTYPE="URL"
        )
        mdr.set("{http://www.w3.org/1999/xlink}href", "url")
        obj = metsrw.SubSection.parse(elem)
        new = obj.serialize(now="2001-02-03T09:10")
        assert new.tag == "{http://www.loc.gov/METS/}techMD"
        assert new.attrib["ID"] == "techMD_42"
        assert new.attrib["CREATED"] == "2016-01-02T03:04:05"
        assert new.attrib["STATUS"] == "original"
        assert len(new.attrib) == 3

    def test_roundtrip_bare(self):
        """It should be able to parse and write out a subsection unchanged."""
        elem = etree.Element("{http://www.loc.gov/METS/}techMD", ID="techMD_42")
        mdr = etree.SubElement(
            elem, "{http://www.loc.gov/METS/}mdRef", MDTYPE="dummy", LOCTYPE="URL"
        )
        mdr.set("{http://www.w3.org/1999/xlink}href", "url")
        obj = metsrw.SubSection.parse(elem)
        new = obj.serialize(now="2001-02-03T09:10")
        assert new.tag == "{http://www.loc.gov/METS/}techMD"
        assert new.attrib["ID"] == "techMD_42"
        assert len(new.attrib) == 1

    def test_roundtrip_group_id(self):
        """It should be able to maintain the GROUPID attribute."""
        elem = etree.Element(
            "{http://www.loc.gov/METS/}dmdSec",
            ID="dmd_1",
            GROUPID="ab3ba63f-3e50-4477-964f-9c5ef2d33dc0",
        )
        mdr = etree.SubElement(
            elem, "{http://www.loc.gov/METS/}mdRef", MDTYPE="dummy", LOCTYPE="URL"
        )
        mdr.set("{http://www.w3.org/1999/xlink}href", "url")
        obj = metsrw.SubSection.parse(elem)
        new = obj.serialize()
        assert new.attrib["GROUPID"] == "ab3ba63f-3e50-4477-964f-9c5ef2d33dc0"

    def test_generate_id_skips_existing_id(self):
        techmd1 = metsrw.SubSection("techMD", [], section_id="techMD_1")
        techmd2 = metsrw.SubSection("techMD", [])
        techmd3 = metsrw.SubSection("techMD", [])

        assert techmd1.id_string != techmd2.id_string
        assert techmd1.id_string != techmd3.id_string
        assert techmd2.id_string != techmd3.id_string

        assert metsrw.SubSection.get_current_id_count("techMD") == 3


class TestMDRef(TestCase):
    """Test MDRef class."""

    def test_create_defaults(self):
        mdref = metsrw.MDRef("path/to/file.txt", "PREMIS:DUMMY", "URL")
        mdreffed = mdref.serialize()

        assert mdreffed.get("LOCTYPE") == "URL"
        assert mdreffed.get(metsrw.lxmlns("xlink") + "href") == "path/to/file.txt"
        assert mdreffed.get("MDTYPE") == "PREMIS:DUMMY"

    def test_create_extra_params(self):
        mdref = metsrw.MDRef(
            target="path/to/file.txt",
            mdtype="OTHER",
            label="Label",
            loctype="OTHER",
            otherloctype="OUTSIDE",
            xptr="xpointer(id('dmdSec_366 dmdSec_367'))",
            othermdtype="METSRIGHTS",
        )
        mdreffed = mdref.serialize()

        assert mdreffed.get("LOCTYPE") == "OTHER"
        assert mdreffed.get("OTHERLOCTYPE") == "OUTSIDE"
        assert mdreffed.get(metsrw.lxmlns("xlink") + "href") == "path/to/file.txt"
        assert mdreffed.get("MDTYPE") == "OTHER"
        assert mdreffed.get("XPTR") == "xpointer(id('dmdSec_366 dmdSec_367'))"
        assert mdreffed.get("OTHERMDTYPE") == "METSRIGHTS"

    def test_create_bad_loctype(self):
        metsrw.MDRef(None, None, loctype="ARK")
        metsrw.MDRef(None, None, loctype="URN")
        metsrw.MDRef(None, None, loctype="URL")
        metsrw.MDRef(None, None, loctype="PURL")
        metsrw.MDRef(None, None, loctype="HANDLE")
        metsrw.MDRef(None, None, loctype="DOI")
        metsrw.MDRef(None, None, loctype="OTHER")
        with pytest.raises(ValueError):
            metsrw.MDRef(None, None, loctype="BAD")

    def test_parse(self):
        # Parses correctly
        good = etree.Element(
            "{http://www.loc.gov/METS/}mdRef", MDTYPE="dummy", LOCTYPE="URL"
        )
        good.set("{http://www.w3.org/1999/xlink}href", "url")
        mdref = metsrw.MDRef.parse(good)
        assert mdref.target == "url"
        assert mdref.mdtype == "dummy"
        assert mdref.loctype == "URL"

    def test_parse_bad_tag_name(self):
        # Wrong tag name
        bad = etree.Element("foo")
        with pytest.raises(metsrw.ParseError) as e:
            metsrw.MDRef.parse(bad)
            assert "METS namespace" in e.value

    def test_parse_no_mdtype(self):
        # No MDTYPE
        bad = etree.Element("{http://www.loc.gov/METS/}mdRef")
        with pytest.raises(metsrw.ParseError) as e:
            metsrw.MDRef.parse(bad)
            assert "MDTYPE" in e.value

    def test_parse_no_href(self):
        # No xlink:href
        bad = etree.Element("{http://www.loc.gov/METS/}mdRef", MDTYPE="PREMIS:DUMMY")
        with pytest.raises(metsrw.ParseError) as e:
            metsrw.MDRef.parse(bad)
            assert "xlink:href" in e.value

    def test_parse_no_loctype(self):
        # No LOCTYPE
        bad = etree.Element("{http://www.loc.gov/METS/}mdRef", MDTYPE="PREMIS:DUMMY")
        bad.set("{http://www.w3.org/1999/xlink}href", "url")
        with pytest.raises(metsrw.ParseError) as e:
            metsrw.MDRef.parse(bad)
            assert "LOCTYPE" in e.value

    def test_url_encoding(self):
        """Good target values are URL-encoded when they show up in xlink:href
        attributes; bad target values raise ``MetsError``.
        """
        mdref = metsrw.MDRef("30_CFLQ_271_13-3-13_1524[1].pdf", "PREMIS:DUMMY", "URL")
        mdreffed = mdref.serialize()
        assert mdreffed.get(metsrw.lxmlns("xlink") + "href") == (
            "30_CFLQ_271_13-3-13_1524%5B1%5D.pdf"
        )
        with pytest.raises(
            metsrw.exceptions.SerializeError, match="is not a valid URL."
        ):
            mdref = metsrw.MDRef(
                "http://foo[bar.com/hello[1].pdf", "PREMIS:DUMMY", "URL"
            )
            mdref.serialize()


class TestMDWrap(TestCase):
    """Test MDWrap class."""

    def test_parse_bad_tag(self):
        """It should assert if the tag name is invalid."""
        bad = etree.Element("foo")
        with pytest.raises(metsrw.ParseError):
            metsrw.MDWrap.parse(bad)

    def test_parse_no_mdtype(self):
        """It should assert if there is no MDTYPE."""
        bad = etree.Element("{http://www.loc.gov/METS/}mdWrap")
        with pytest.raises(metsrw.ParseError):
            metsrw.MDWrap.parse(bad)

    def test_parse_no_children(self):
        """It should assert if there are no children."""
        # mdWrap has no children
        bad = etree.Element("{http://www.loc.gov/METS/}mdWrap", MDTYPE="dummy")
        with pytest.raises(metsrw.ParseError):
            metsrw.MDWrap.parse(bad)
        # xmlData has no children
        bad = etree.Element("{http://www.loc.gov/METS/}mdWrap", MDTYPE="dummy")
        etree.SubElement(bad, "{http://www.loc.gov/METS/}xmlData")
        with pytest.raises(metsrw.ParseError):
            metsrw.MDWrap.parse(bad)

    def test_parse_success(self):
        """It should correctly parse a valid mdWrap."""
        # Parses correctly
        good = etree.Element(
            "{http://www.loc.gov/METS/}mdWrap", MDTYPE="OTHER", OTHERMDTYPE="SYSTEM"
        )
        xmldata = etree.SubElement(good, "{http://www.loc.gov/METS/}xmlData")
        etree.SubElement(xmldata, "foo")
        mdwrap = metsrw.MDWrap.parse(good)
        assert mdwrap.mdtype == "OTHER"
        assert mdwrap.othermdtype == "SYSTEM"
        assert etree.tostring(mdwrap.document) == etree.tostring(etree.Element("foo"))

    def test_serialize_defaults(self):
        mdwrap = metsrw.MDWrap("<foo/>", "PREMIS:DUMMY")
        elem = mdwrap.serialize()

        assert elem.tag == "{http://www.loc.gov/METS/}mdWrap"
        assert elem.attrib["MDTYPE"] == "PREMIS:DUMMY"
        assert len(elem.attrib) == 1
        assert elem[0].tag == "{http://www.loc.gov/METS/}xmlData"
        assert len(elem[0].attrib) == 0
        assert elem[0][0].tag == "foo"

    def test_serialize_params(self):
        mdwrap = metsrw.MDWrap(
            etree.Element("foo"), mdtype="OTHER", othermdtype="SYSTEM"
        )
        elem = mdwrap.serialize()

        assert elem.tag == "{http://www.loc.gov/METS/}mdWrap"
        assert elem.attrib["MDTYPE"] == "OTHER"
        assert elem.attrib["OTHERMDTYPE"] == "SYSTEM"
        assert len(elem.attrib) == 2
        assert elem[0].tag == "{http://www.loc.gov/METS/}xmlData"
        assert len(elem[0].attrib) == 0
        assert elem[0][0].tag == "foo"

    def test_roundtrip(self):
        elem = etree.Element(
            "{http://www.loc.gov/METS/}mdWrap", MDTYPE="OTHER", OTHERMDTYPE="SYSTEM"
        )
        xmldata = etree.SubElement(elem, "{http://www.loc.gov/METS/}xmlData")
        etree.SubElement(xmldata, "foo")

        mdwrap = metsrw.MDWrap.parse(elem)
        elem = mdwrap.serialize()

        assert elem.tag == "{http://www.loc.gov/METS/}mdWrap"
        assert elem.attrib["MDTYPE"] == "OTHER"
        assert elem.attrib["OTHERMDTYPE"] == "SYSTEM"
        assert len(elem.attrib) == 2
        assert elem[0].tag == "{http://www.loc.gov/METS/}xmlData"
        assert len(elem[0].attrib) == 0
        assert elem[0][0].tag == "foo"

    def test_url_decoding(self):
        good = etree.Element(
            "{http://www.loc.gov/METS/}mdRef", MDTYPE="dummy", LOCTYPE="URL"
        )
        good.set(
            "{http://www.w3.org/1999/xlink}href", "30_CFLQ_271_13-3-13_1524%5B1%5D.pdf"
        )
        mdref = metsrw.MDRef.parse(good)
        assert mdref.target == "30_CFLQ_271_13-3-13_1524[1].pdf"
        with pytest.raises(metsrw.exceptions.ParseError, match="is not a valid URL"):
            bad = etree.Element(
                "{http://www.loc.gov/METS/}mdRef", MDTYPE="dummy", LOCTYPE="URL"
            )
            bad.set(
                "{http://www.w3.org/1999/xlink}href", "http://foo[bar.com/hello[1].pdf"
            )
            metsrw.MDRef.parse(bad)



from unittest import TestCase

import metsrw
import metsrw.plugins.premisrw as premisrw

from .constants import EX_AGT_1
from .constants import EX_AGT_2
from .constants import EX_COMPR_EVT
from .constants import EX_PTR_AIP_SUBTYPE
from .constants import EX_PTR_DATE_CREATED_BY_APPLICATION
from .constants import EX_PTR_FORMAT_NAME
from .constants import EX_PTR_FORMAT_REGISTRY_KEY
from .constants import EX_PTR_IDENTIFIER_VALUE
from .constants import EX_PTR_MESSAGE_DIGEST
from .constants import EX_PTR_MESSAGE_DIGEST_ALGORITHM
from .constants import EX_PTR_PACKAGE_TYPE
from .constants import EX_PTR_PATH
from .constants import EX_PTR_SIZE
from .constants import EX_PTR_XSI_TYPE


class BetterPREMISObject:
    """A dummy PREMIS object class that can be used to meet the dependency of
    ``FSEntry``.
    """

    def __init__(self, tree):
        self.tree = tree

    @classmethod
    def fromtree(cls, tree):
        return cls(tree)

    def serialize(self):
        return self.tree


class TestDependencyInjection(TestCase):
    """Test dependency injection of metadata functionality into the
    FSEntry class.
    """

    def test_assertion_creators(self):
        """Tests that the assertion creators ``has_methods`` and
        ``has_class_methods`` behave correctly.

        Note the false positives below. In python 3 it is easy to distinguish
        regular methods from class methods in a class but to do it in a 2/3
        compatible way seems like too much introspection bother right now. The
        more difficult issue would seem to be distinguishing regular methods
        from static methods on a class. However, these are acceptable flaws if
        we are mainly concerned with asserting the existence of a) specific
        class methods on classes and b) specific methods (of any type) on
        classes.
        """

        class C:
            @classmethod
            def cm(cls):
                pass

            @staticmethod
            def sm():
                pass

            def rm(self):
                pass

        c = C()

        class X:
            pass

        x = X()

        has_rm_method = metsrw.has_methods("rm")
        has_cm_method = metsrw.has_methods("cm")
        has_sm_method = metsrw.has_methods("sm")
        has_xm_method = metsrw.has_methods("xm")

        has_rm_class_method = metsrw.has_class_methods("rm")
        has_cm_class_method = metsrw.has_class_methods("cm")
        has_sm_class_method = metsrw.has_class_methods("sm")
        has_xm_class_method = metsrw.has_class_methods("xm")

        # C should have classmethod 'cm' and regular method 'rm'
        assert has_rm_method(C) is True
        assert has_cm_method(C) is True  # false positive
        assert has_sm_method(C) is True  # false positive
        assert has_xm_method(C) is False
        assert has_rm_class_method(C) is False
        assert has_cm_class_method(C) is True
        assert has_sm_class_method(C) is False
        assert has_xm_class_method(C) is False

        # c should have classmethod 'cm' and regular method 'rm'
        assert has_rm_method(c) is True
        assert has_cm_method(c) is False
        assert has_sm_method(c) is False
        assert has_xm_method(c) is False
        assert has_rm_class_method(c) is False
        assert has_cm_class_method(c) is True
        assert has_sm_class_method(c) is False
        assert has_xm_class_method(c) is False

        # X should have nothing
        assert has_rm_method(X) is False
        assert has_cm_method(X) is False
        assert has_sm_method(X) is False
        assert has_xm_method(X) is False
        assert has_rm_class_method(X) is False
        assert has_cm_class_method(X) is False
        assert has_sm_class_method(X) is False
        assert has_xm_class_method(X) is False

        # x should have nothing
        assert has_rm_method(x) is False
        assert has_cm_method(x) is False
        assert has_sm_method(x) is False
        assert has_xm_method(x) is False
        assert has_rm_class_method(x) is False
        assert has_cm_class_method(x) is False
        assert has_sm_class_method(x) is False
        assert has_xm_class_method(x) is False

    def test_dependency_injection(self):
        """Test the dependency injection (DI) infrastructure for metsrw plugins.

        - client: metsrw.FSEntry
        - services: classes for reading and writing metadata elements, e.g.,
          the PREMISObject class of metsrw.plugins.premisrw or other classes
          exposing the same interface.
        - injector: this test code or the code in metsrw/di.py which calls
          ``provide`` on the ``feature_broker`` singleton.

        The ``FSEntry`` class declares its dependency on the class attributes
        ``premis_object_class``, ``premis_event_class``, and
        ``premis_agent_class`` and further requires that these return classes
        with ``fromtree`` and ``serialize`` methods::

            >>> premis_object_class = Dependency(
            ...     has_methods('serialize'),
            ...     has_class_methods('fromtree'),
            ...     is_class)

        """

        # Clear the feature broker and then register/provide the premisrw
        # plugin classes (services) with the feature broker.
        feature_broker = metsrw.feature_broker
        assert len(feature_broker) == 4
        feature_broker.clear()
        assert not feature_broker
        feature_broker.provide("premis_object_class", premisrw.PREMISObject)
        feature_broker.provide("premis_event_class", premisrw.PREMISEvent)
        feature_broker.provide("premis_agent_class", premisrw.PREMISAgent)
        feature_broker.provide("premis_rights_class", premisrw.PREMISRights)
        assert len(feature_broker) == 4

        # Create premisrw instances.
        compression_premis_event = premisrw.PREMISEvent(data=EX_COMPR_EVT)
        premis_events = [compression_premis_event]
        premis_agents = [premisrw.PREMISAgent(data=x) for x in [EX_AGT_1, EX_AGT_2]]
        (
            _,
            compression_program_version,
            archive_tool,
        ) = compression_premis_event.compression_details
        premis_object = premisrw.PREMISObject(
            xsi_type=EX_PTR_XSI_TYPE,
            identifier_value=EX_PTR_IDENTIFIER_VALUE,
            message_digest_algorithm=EX_PTR_MESSAGE_DIGEST_ALGORITHM,
            message_digest=EX_PTR_MESSAGE_DIGEST,
            size=EX_PTR_SIZE,
            format_name=EX_PTR_FORMAT_NAME,
            format_registry_key=EX_PTR_FORMAT_REGISTRY_KEY,
            creating_application_name=archive_tool,
            creating_application_version=compression_program_version,
            date_created_by_application=EX_PTR_DATE_CREATED_BY_APPLICATION,
        )
        transform_files = compression_premis_event.get_decompression_transform_files()

        # Create metsrw ``METSDocument`` and ``FSEntry`` instances.
        mets_doc = metsrw.METSDocument()
        fs_entry = metsrw.FSEntry(
            path=EX_PTR_PATH,
            file_uuid=EX_PTR_IDENTIFIER_VALUE,
            use=EX_PTR_PACKAGE_TYPE,
            type=EX_PTR_PACKAGE_TYPE,
            transform_files=transform_files,
            mets_div_type=EX_PTR_AIP_SUBTYPE,
        )
        mets_doc.append_file(fs_entry)

        # Use the ``add_premis_...`` methods to add the PREMIS metadata
        # elements to the ``FSEntry`` instance. This will assert that each
        # PREMIS instance is of the correct type (e.g., that ``premis_object``
        # is an instance of ``FSEntry().premis_object_class``) and will call the
        # instance's ``serialize`` method and incorporate the resulting
        # ``lxml.etree._ElementTree`` instance into the ``FSEntry`` instance
        # appropriately.
        fs_entry.add_premis_object(premis_object)
        for premis_event in premis_events:
            fs_entry.add_premis_event(premis_event)
        for premis_agent in premis_agents:
            fs_entry.add_premis_agent(premis_agent)

        # Assert that the instances returned by the
        # ``FSEntry().get_premis_...`` methods are of the anticipated type.
        new_premis_agents = fs_entry.get_premis_agents()
        for new_premis_agent in new_premis_agents:
            assert isinstance(new_premis_agent, premisrw.PREMISAgent)
            assert new_premis_agent in premis_agents
            assert id(new_premis_agent) not in [id(pa) for pa in premis_agents]
        new_premis_events = fs_entry.get_premis_events()
        for new_premis_event in new_premis_events:
            assert isinstance(new_premis_event, premisrw.PREMISEvent)
            assert new_premis_event in premis_events
            assert id(new_premis_event) not in [id(pa) for pa in premis_events]
        new_premis_objects = fs_entry.get_premis_objects()
        for new_premis_object in new_premis_objects:
            assert isinstance(new_premis_object, premisrw.PREMISObject)
            assert new_premis_object == premis_object
            assert id(new_premis_object) is not premis_object

        # Assert that the resulting mets XML contains a
        # premis:objectIdentifierValue in the anticipated location in the
        # structure with the anticipated value.
        mets_doc_el = mets_doc.serialize()
        xpath = (
            'mets:amdSec/mets:techMD/mets:mdWrap[@MDTYPE="PREMIS:OBJECT"]'
            "/mets:xmlData/premis:object/premis:objectIdentifier/"
            "premis:objectIdentifierValue"
        )
        a = mets_doc_el.find(xpath, namespaces=metsrw.NAMESPACES)
        assert a.text == EX_PTR_IDENTIFIER_VALUE

        # Now change the feature broker so that ``FSEntry``'s dependency on a
        # ``premis_object_class`` class attribute is being fulfilled by a new
        # class: ``BetterPREMISObject``.
        feature_broker.provide("premis_object_class", BetterPREMISObject)

        # Now create a new PREMIS object
        premis_object_tree = premis_object.serialize()
        better_premis_object = BetterPREMISObject.fromtree(premis_object_tree)

        # And re-create the ``METSDocument`` and ``FSEntry`` instances.
        mets_doc = metsrw.METSDocument()
        fs_entry = metsrw.FSEntry(
            path=EX_PTR_PATH,
            file_uuid=EX_PTR_IDENTIFIER_VALUE,
            use=EX_PTR_PACKAGE_TYPE,
            type=EX_PTR_PACKAGE_TYPE,
            transform_files=transform_files,
            mets_div_type=EX_PTR_AIP_SUBTYPE,
        )
        mets_doc.append_file(fs_entry)

        # Add the PREMIS metadata again, but this time use the instance of
        # ``BetterPREMISObject``.
        fs_entry.add_premis_object(better_premis_object)
        for premis_event in premis_events:
            fs_entry.add_premis_event(premis_event)
        for premis_agent in premis_agents:
            fs_entry.add_premis_agent(premis_agent)

        # Assert that the instances returned by the
        # ``FSEntry().get_premis_...`` methods are of the anticipated type.
        new_premis_objects = fs_entry.get_premis_objects()
        for new_premis_object in new_premis_objects:
            assert isinstance(new_premis_object, BetterPREMISObject)

        # Make sure we can still find the PREMIS object id value.
        mets_doc_el = mets_doc.serialize()
        assert (
            mets_doc_el.find(xpath, namespaces=metsrw.NAMESPACES).text
            == EX_PTR_IDENTIFIER_VALUE
        )

        # Reset the feature broker to its default state so subsequent tests
        # don't break.
        metsrw.set_feature_broker_to_default_state(feature_broker)



import pytest

import metsrw


@pytest.mark.parametrize(
    "fixture_path",
    [
        "fixtures/mets_structmap_imported.xml",
        "fixtures/mets_structmap_arranged_sip.xml",
    ],
    ids=["mets_with_structmap_imported_from_xml", "mets_from_arranged_sip"],
)
def test_custom_struct_maps_are_serialized(fixture_path):
    """
    Logical structMaps other than the normative one should be parsed and
    serialized correctly.
    """
    mets = metsrw.METSDocument.fromfile(fixture_path)
    assert len(mets._custom_structmaps) == 1

    result = mets.serialize()
    custom_struct_maps = [
        e
        for e in result.findall(
            'mets:structMap[@TYPE="logical"]',
            namespaces=metsrw.NAMESPACES,
        )
        if e.attrib.get("LABEL") != "Normative Directory Structure"
    ]
    assert len(custom_struct_maps) == 1



import uuid
from unittest import TestCase

import pytest

import metsrw


class TestFSEntry(TestCase):
    """Test FSEntry class."""

    def test_path_is_str(self):
        """It should store the ``path`` as a str."""
        sample = "💜🎑💜"
        assert isinstance(
            metsrw.FSEntry(sample.encode("utf-8"), type="Directory").path,
            str,
        )

    def test_create_invalid_checksum_type(self):
        """It should only accept METS valid checksum types."""
        metsrw.FSEntry("file[1].txt", checksumtype="Adler-32", checksum="dummy")
        metsrw.FSEntry("file[1].txt", checksumtype="CRC32", checksum="dummy")
        metsrw.FSEntry("file[1].txt", checksumtype="HAVAL", checksum="dummy")
        metsrw.FSEntry(
            "file[1].txt",
            checksumtype="MD5",
            checksum="daa05c683a4913b268653f7a7e36a5b4",
        )
        metsrw.FSEntry("file[1].txt", checksumtype="MNP", checksum="dummy")
        metsrw.FSEntry("file[1].txt", checksumtype="SHA-1", checksum="dummy")
        metsrw.FSEntry("file[1].txt", checksumtype="SHA-256", checksum="dummy")
        metsrw.FSEntry("file[1].txt", checksumtype="SHA-384", checksum="dummy")
        metsrw.FSEntry("file[1].txt", checksumtype="SHA-512", checksum="dummy")
        metsrw.FSEntry("file[1].txt", checksumtype="TIGER WHIRLPOOL", checksum="dummy")
        with pytest.raises(ValueError):
            metsrw.FSEntry("file[1].txt", checksumtype="DNE", checksum="dummy")

    def test_create_checksum_and_checksumtype(self):
        with pytest.raises(ValueError):
            metsrw.FSEntry("file[1].txt", checksum="daa05c683a4913b268653f7a7e36a5b4")
        with pytest.raises(ValueError):
            metsrw.FSEntry("file[1].txt", checksumtype="MD5")

    def test_file_id_directory(self):
        """It should have no file ID."""
        d = metsrw.FSEntry("dir", type="Directory")
        assert d.file_id() is None

    def test_file_id_no_uuid(self):
        """It should raise an exception with no file UUID."""
        f = metsrw.FSEntry("level1.txt")
        with pytest.raises(metsrw.MetsError):
            f.file_id()

    def test_file_id_success(self):
        """It should return a file ID."""
        file_uuid = str(uuid.uuid4())
        f = metsrw.FSEntry("level1.txt", file_uuid=file_uuid)
        assert f.file_id() == "file-" + file_uuid

    def test_aip_file_id(self):
        fsentry = metsrw.FSEntry(
            file_uuid="9b9f129c-8062-471b-a009-9ee0ad655f08",
            type="Archival Information Package",
            path="/tmp/example-1-9b9f129c-8062-471b-a009-9ee0ad655f08.7z",
        )
        assert (
            fsentry.file_id() == "file-example-1-9b9f129c-8062-471b-a009-9ee0ad655f08"
        )

    def test_group_id_no_uuid(self):
        """It should return no group ID."""
        f = metsrw.FSEntry("level1.txt")
        assert f.group_id() is None

    def test_group_id_uuid(self):
        """It should return a group ID."""
        file_uuid = str(uuid.uuid4())
        f = metsrw.FSEntry("level1.txt", file_uuid=file_uuid)
        assert f.group_id() == "Group-" + file_uuid

    def test_group_id_derived(self):
        """It should return the group ID for the derived from file."""
        file_uuid = str(uuid.uuid4())
        f = metsrw.FSEntry("level1.txt", file_uuid=file_uuid)
        derived = metsrw.FSEntry(
            "level3.txt", file_uuid=str(uuid.uuid4()), derived_from=f
        )
        assert derived.group_id() == "Group-" + file_uuid
        assert derived.group_id() == f.group_id()

    def test_admids(self):
        """It should return 0 or 1 amdSecs."""
        f = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))
        assert len(f.admids) == 0
        f.add_premis_object("<premis>object</premis>")
        assert len(f.admids) == 1
        f.add_premis_event("<premis>event</premis>")
        # Can only have one amdSec
        assert len(f.admids) == 1

    def test_dmdids(self):
        """It should return a DMDID for each dmdSec."""
        f = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))
        assert len(f.dmdids) == 0
        f.add_dublin_core("<dc />")
        assert len(f.dmdids) == 1

    def test_add_metadata_to_fsentry(self):
        f1 = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))
        f1.add_dublin_core("<dc />")
        assert f1.dmdsecs
        assert len(f1.dmdsecs) == 1
        assert f1.dmdsecs[0].subsection == "dmdSec"
        assert f1.dmdsecs[0].contents.mdtype == "DC"

        # Can only have 1 amdSec, so subsequent subsections are children of
        # AMDSec
        f1.add_premis_object("<premis>object</premis>")
        assert f1.amdsecs
        assert f1.amdsecs[0].subsections
        assert f1.amdsecs[0].subsections[0].subsection == "techMD"
        assert f1.amdsecs[0].subsections[0].contents.mdtype == "PREMIS:OBJECT"

        f1.add_premis_event("<premis>event</premis>")
        assert f1.amdsecs[0].subsections[1].subsection == "digiprovMD"
        assert f1.amdsecs[0].subsections[1].contents.mdtype == "PREMIS:EVENT"

        f1.add_premis_agent("<premis>agent</premis>")
        assert f1.amdsecs[0].subsections[2].subsection == "digiprovMD"
        assert f1.amdsecs[0].subsections[2].contents.mdtype == "PREMIS:AGENT"

        f1.add_premis_rights("<premis>rights</premis>")
        assert f1.amdsecs[0].subsections[3].subsection == "rightsMD"
        assert f1.amdsecs[0].subsections[3].contents.mdtype == "PREMIS:RIGHTS"

        assert len(f1.amdsecs[0].subsections) == 4

    def test_dmdsec_management(self):
        """Test addition, check and deletion of dmdSecs."""
        file = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))
        assert file.dmdsecs == []
        assert file.dmdsecs_by_mdtype == {}
        assert file.has_dmdsec("DC") is False
        assert file.has_dmdsec("OTHER_CUSTOM") is False
        file.add_dmdsec("<dc/>", "DC")
        assert len(file.dmdsecs) == 1
        assert len(file.dmdsecs_by_mdtype["DC"]) == 1
        assert file.has_dmdsec("DC") is True
        assert file.dmdsecs[0].status == "original"
        file.add_dmdsec("<dc/>", "DC", status="update")
        assert len(file.dmdsecs) == 2
        assert len(file.dmdsecs_by_mdtype["DC"]) == 2
        assert file.dmdsecs[0].status == "original-superseded"
        assert file.dmdsecs[1].status == "update"
        assert file.dmdsecs[0].group_id == file.dmdsecs[1].group_id
        file.add_dmdsec("<dc/>", "DC", status="update")
        assert len(file.dmdsecs) == 3
        assert len(file.dmdsecs_by_mdtype["DC"]) == 3
        assert file.dmdsecs[1].status == "update-superseded"
        assert file.dmdsecs[2].status == "update"
        assert file.dmdsecs[0].group_id == file.dmdsecs[1].group_id
        assert file.dmdsecs[0].group_id == file.dmdsecs[2].group_id
        file.delete_dmdsec("DC")
        assert len(file.dmdsecs) == 3
        assert len(file.dmdsecs_by_mdtype["DC"]) == 3
        assert file.dmdsecs[2].status == "deleted"
        file.add_dmdsec("<custom/>", "OTHER", othermdtype="CUSTOM")
        assert len(file.dmdsecs) == 4
        assert len(file.dmdsecs_by_mdtype["DC"]) == 3
        assert len(file.dmdsecs_by_mdtype["OTHER_CUSTOM"]) == 1
        assert file.has_dmdsec("OTHER", othermdtype="CUSTOM") is True
        assert file.dmdsecs[3].status == "original"
        file.add_dmdsec("<custom/>", "OTHER", othermdtype="CUSTOM", status="update")
        assert len(file.dmdsecs) == 5
        assert len(file.dmdsecs_by_mdtype["DC"]) == 3
        assert len(file.dmdsecs_by_mdtype["OTHER_CUSTOM"]) == 2
        assert file.dmdsecs[3].status == "original-superseded"
        assert file.dmdsecs[4].status == "update"
        assert file.dmdsecs[3].group_id == file.dmdsecs[4].group_id
        file.add_dmdsec("<custom/>", "OTHER", othermdtype="CUSTOM", status="update")
        assert len(file.dmdsecs) == 6
        assert len(file.dmdsecs_by_mdtype["DC"]) == 3
        assert len(file.dmdsecs_by_mdtype["OTHER_CUSTOM"]) == 3
        assert file.dmdsecs[4].status == "update-superseded"
        assert file.dmdsecs[5].status == "update"
        assert file.dmdsecs[3].group_id == file.dmdsecs[4].group_id
        assert file.dmdsecs[3].group_id == file.dmdsecs[5].group_id
        file.delete_dmdsec("OTHER", othermdtype="CUSTOM")
        assert len(file.dmdsecs) == 6
        assert len(file.dmdsecs_by_mdtype["DC"]) == 3
        assert len(file.dmdsecs_by_mdtype["OTHER_CUSTOM"]) == 3
        assert file.dmdsecs[5].status == "deleted"

    def test_add_child(self):
        """
        It should add a new entry to the children list.
        It should add a parent link.
        It should handle duplicates.
        """
        d = metsrw.FSEntry("dir", type="Directory")
        f = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))

        d.add_child(f)
        assert f in d.children
        assert len(d.children) == 1
        assert f.parent is d

        d.add_child(f)
        assert f in d.children
        assert len(d.children) == 1
        assert f.parent is d

        with pytest.raises(ValueError):
            f.add_child(d)

    def test_remove_child(self):
        """
        It should remove the child from the parent's children list.
        It should remove the parent from the child's parent link.
        """
        d = metsrw.FSEntry("dir", type="Directory")
        f1 = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))
        f2 = metsrw.FSEntry("file2.txt", file_uuid=str(uuid.uuid4()))
        d.add_child(f1)
        d.add_child(f2)
        assert f1 in d.children
        assert f1.parent is d
        assert len(d.children) == 2

        d.remove_child(f1)

        assert f1 not in d.children
        assert f1.parent is None
        assert len(d.children) == 1

    def test_serialize_filesec_basic(self):
        """
        It should produce a mets:file element.
        It should have an ID attribute.
        It should not have ADMIDs.
        It should have a child mets:FLocat element with the path.
        """
        f = metsrw.FSEntry(
            "file[1].txt",
            file_uuid=str(uuid.uuid4()),
            checksumtype="MD5",
            checksum="daa05c683a4913b268653f7a7e36a5b4",
        )
        el = f.serialize_filesec()
        assert el.tag == "{http://www.loc.gov/METS/}file"
        assert el.attrib["ID"].startswith("file-")
        assert el.attrib["CHECKSUM"] == "daa05c683a4913b268653f7a7e36a5b4"
        assert el.attrib["CHECKSUMTYPE"] == "MD5"
        assert el.attrib.get("ADMID") is None
        assert len(el) == 1
        assert el[0].tag == "{http://www.loc.gov/METS/}FLocat"
        assert el[0].attrib["LOCTYPE"] == "OTHER"
        assert el[0].attrib["OTHERLOCTYPE"] == "SYSTEM"
        assert el[0].attrib["{http://www.w3.org/1999/xlink}href"] == "file%5B1%5D.txt"

    def test_serialize_filesec_metadata(self):
        """
        It should produce a mets:file element.
        It should have an ID attribute.
        It should have one ADMID.
        It should have a child mets:FLocat element with the path.
        """
        f = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))
        f.add_premis_object("<premis>object</premis>")
        el = f.serialize_filesec()
        assert el.tag == "{http://www.loc.gov/METS/}file"
        assert el.attrib["ID"].startswith("file-")
        assert len(el.attrib["ADMID"].split()) == 1
        assert len(el) == 1
        assert el[0].tag == "{http://www.loc.gov/METS/}FLocat"
        assert el[0].attrib["LOCTYPE"] == "OTHER"
        assert el[0].attrib["OTHERLOCTYPE"] == "SYSTEM"
        assert el[0].attrib["{http://www.w3.org/1999/xlink}href"] == "file%5B1%5D.txt"

    def test_serialize_filesec_not_item(self):
        """
        It should not produce a mets:file element.
        """
        f = metsrw.FSEntry("file[1].txt", type="Directory", file_uuid=str(uuid.uuid4()))
        el = f.serialize_filesec()
        assert el is None

    def test_serialize_filesec_no_use(self):
        """
        It should not produce a mets:file element.
        """
        f = metsrw.FSEntry("file[1].txt", use=None, file_uuid=str(uuid.uuid4()))
        el = f.serialize_filesec()
        assert el is None

    def test_serialize_filesec_transform_files(self):
        """
        It should produce a mets:file element.
        It should have a child mets:FLocat element.
        It should have a child mets:transformFile element.
        """
        transform_files = [
            {"type": "decryption", "order": 1, "algorithm": "GPG", "key": "somekey"}
        ]
        f = metsrw.FSEntry(
            "file[1].txt", file_uuid=str(uuid.uuid4()), transform_files=transform_files
        )
        file_element = f.serialize_filesec()
        assert file_element.tag == "{http://www.loc.gov/METS/}file"
        assert file_element[0].tag == "{http://www.loc.gov/METS/}FLocat"
        assert file_element[1].tag == "{http://www.loc.gov/METS/}transformFile"
        assert file_element[1].attrib["TRANSFORMTYPE"] == "decryption"
        assert file_element[1].attrib["TRANSFORMORDER"] == "1"
        assert file_element[1].attrib["TRANSFORMALGORITHM"] == "GPG"
        assert file_element[1].attrib["TRANSFORMKEY"] == "somekey"

    def test_serialize_filesec_no_path(self):
        """
        It should produce a mets:file element.
        It should not have a child mets:FLocat.
        """
        file_uuid = str(uuid.uuid4())
        f = metsrw.FSEntry(file_uuid=file_uuid, use="deletion")
        el = f.serialize_filesec()
        assert el.tag == "{http://www.loc.gov/METS/}file"
        assert el.attrib["ID"] == "file-" + file_uuid
        assert el.attrib["GROUPID"] == "Group-" + file_uuid
        assert len(el.attrib) == 2
        assert len(el) == 0

    def test_serialize_structmap_file(self):
        """
        It should produce a mets:div element.
        It should have a TYPE and LABEL.
        It should have a child mets:fptr element with FILEID.
        """
        f = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))
        f.add_dublin_core("<dc />")
        el = f.serialize_structmap(recurse=False)
        assert el.tag == "{http://www.loc.gov/METS/}div"
        assert el.attrib["TYPE"] == "Item"
        assert el.attrib["LABEL"] == "file[1].txt"
        assert len(el.attrib["DMDID"].split()) == 1
        assert len(el) == 1
        assert el[0].tag == "{http://www.loc.gov/METS/}fptr"
        assert el[0].attrib["FILEID"].startswith("file-")

    def test_serialize_structmap_no_recurse(self):
        """
        It should produce a mets:div element.
        It should have a TYPE and LABEL.
        It should not have children.
        """
        d = metsrw.FSEntry("dir", type="Directory")
        f = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))
        d.add_child(f)
        el = d.serialize_structmap(recurse=False)
        assert el.tag == "{http://www.loc.gov/METS/}div"
        assert el.attrib["TYPE"] == "Directory"
        assert el.attrib["LABEL"] == "dir"
        assert len(el) == 0

    def test_serialize_structmap_recurse(self):
        """
        It should produce a mets:div element.
        It should have a TYPE and LABEL.
        It should have a child mets:div with the file.
        """
        d = metsrw.FSEntry("dir", type="Directory")
        f = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))
        d.add_child(f)
        el = d.serialize_structmap(recurse=True)
        assert el.tag == "{http://www.loc.gov/METS/}div"
        assert el.attrib["TYPE"] == "Directory"
        assert el.attrib["LABEL"] == "dir"
        assert len(el) == 1
        assert el[0].tag == "{http://www.loc.gov/METS/}div"
        assert el[0].attrib["TYPE"] == "Item"
        assert el[0].attrib["LABEL"] == "file[1].txt"
        assert len(el[0]) == 1
        assert el[0][0].tag == "{http://www.loc.gov/METS/}fptr"
        assert el[0][0].attrib["FILEID"].startswith("file-")

    def test_serialize_structmap_no_label(self):
        """It should return None."""
        f = metsrw.FSEntry()
        el = f.serialize_structmap(recurse=False)
        assert el is None

    def test_serialize_structmap_child_empty(self):
        """It should handle children with no structMap entry."""
        d = metsrw.FSEntry("dir", type="Directory")
        f = metsrw.FSEntry(use="deletion", file_uuid=str(uuid.uuid4()))
        d.add_child(f)
        el = d.serialize_structmap(recurse=True)
        assert el.tag == "{http://www.loc.gov/METS/}div"
        assert el.attrib["TYPE"] == "Directory"
        assert el.attrib["LABEL"] == "dir"
        assert len(el.attrib) == 2
        assert len(el) == 0

    def test_serialize_structmap_directory_admid(self):
        """It should add the ADMID attribute to directories with amdsecs."""
        d = metsrw.FSEntry("objects", type="Directory")
        f = metsrw.FSEntry("file.txt", file_uuid=str(uuid.uuid4()))
        d.add_child(f)
        d.add_premis_object("<premis>object</premis>")
        el = d.serialize_structmap()
        assert el.attrib.get("ADMID") is not None

    def test_is_empty_dir(self):
        """It should be able to determine whether it is an empty directory."""
        r = metsrw.FSEntry("dir", type="Directory")
        d1 = metsrw.FSEntry("dir", type="Directory")
        d2 = metsrw.FSEntry("dir", type="Directory")
        d1a = metsrw.FSEntry("dir", type="Directory")
        d2a = metsrw.FSEntry("dir", type="Directory")
        d2b = metsrw.FSEntry("dir", type="Directory")
        f = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))
        r.add_child(d1)
        r.add_child(d2)
        d1.add_child(d1a)
        d2.add_child(d2a)
        d2.add_child(d2b)
        d1a.add_child(f)

        assert d2a.is_empty_dir
        assert not d2a.children
        assert not d1a.is_empty_dir
        assert len(d1a.children) == 1
        assert not d1.is_empty_dir
        assert not r.is_empty_dir
        assert not f.is_empty_dir
        # Directory d2 is an empty directory because it contains nothing but
        # empty directories.
        assert d2.is_empty_dir
        assert len(d2.children) == 2

    def test_get_path(self):
        """It should be able to get or generate a relative path."""
        sip = metsrw.FSEntry(label="sip", type="Directory")
        objects = metsrw.FSEntry(label="objects", type="Directory")
        directory = metsrw.FSEntry(label="directory", type="Directory")
        file = metsrw.FSEntry("objects/directory/file.txt")
        sip.add_child(objects).add_child(directory).add_child(file)

        assert sip.get_path() is None
        assert objects.get_path() == "objects"
        assert directory.get_path() == "objects/directory"
        assert file.get_path() == "objects/directory/file.txt"

    def test_get_path_error(self):
        """It should raise AttributeError is an entry doesn't have path nor label."""
        with pytest.raises(AttributeError):
            metsrw.FSEntry().get_path()



import datetime
import filecmp
import io
import os
import tempfile
import uuid
from unittest import TestCase
from unittest import mock

import pytest
from lxml import etree
from lxml.builder import ElementMaker

import metsrw


class TestMETSDocument(TestCase):
    """Test METSDocument class."""

    def test_fromfile(self):
        parser = etree.XMLParser(remove_blank_text=True)
        root = etree.parse("fixtures/complete_mets.xml", parser=parser)
        mw = metsrw.METSDocument.fromfile("fixtures/complete_mets.xml")
        assert isinstance(mw.tree, etree._ElementTree)
        assert etree.tostring(mw.tree) == etree.tostring(root)

    def test_fromstring(self):
        parser = etree.XMLParser(remove_blank_text=True)
        root = etree.parse("fixtures/complete_mets.xml", parser=parser)
        with open("fixtures/complete_mets.xml", "rb") as f:
            metsstring = f.read()
        mw = metsrw.METSDocument.fromstring(metsstring)
        assert isinstance(mw.tree, etree._ElementTree)
        assert etree.tostring(mw.tree) == etree.tostring(root)

    def test_fromtree(self):
        root = etree.parse("fixtures/complete_mets.xml")
        mw = metsrw.METSDocument.fromtree(root)
        assert isinstance(mw.tree, etree._ElementTree)
        assert etree.tostring(mw.tree) == etree.tostring(root)

    def test_parse(self):
        """
        It should set the correct createdate.
        It should create FSEntrys for every file and directory.
        It should associate amdSec and dmdSec with the FSEntry.
        It should associated derived files.
        """
        mw = metsrw.METSDocument()
        parser = etree.XMLParser(remove_blank_text=True)
        root = etree.parse("fixtures/complete_mets.xml", parser=parser)
        mw.tree = root
        mw._parse_tree()
        assert mw.createdate == "2014-07-23T21:48:33"
        assert len(mw.all_files()) == 14
        assert (
            mw.get_file(
                type="Directory", label="csv-48accdf3-e425-4874-aad3-67ade019a214"
            )
            is not None
        )
        parent = mw.get_file(type="Directory", label="objects")
        assert parent is not None
        f = mw.get_file(type="Item", label="Landing_zone.jpg")
        assert f is not None
        assert f.path == "objects/Landing_zone.jpg"
        assert f.use == "original"
        assert f.parent == parent
        assert f.children == []
        assert f.file_uuid == "ab5c67fc-8f80-4e46-9f20-8d5ae29c43f2"
        assert f.derived_from is None
        assert f.admids == ["amdSec_1"]
        assert f.dmdids == ["dmdSec_1"]
        assert f.file_id() == "file-ab5c67fc-8f80-4e46-9f20-8d5ae29c43f2"
        assert f.group_id() == "Group-ab5c67fc-8f80-4e46-9f20-8d5ae29c43f2"
        f = mw.get_file(
            type="Item", label="Landing_zone-fc33fc0e-40ef-4ad9-ba52-860368e8ce5a.tif"
        )
        assert f is not None
        assert f.path == "objects/Landing_zone-fc33fc0e-40ef-4ad9-ba52-860368e8ce5a.tif"
        assert f.use == "preservation"
        assert f.parent == parent
        assert f.children == []
        assert f.file_uuid == "e284d015-cfb0-45dd-961d-512bf0f47cf6"
        assert f.derived_from == mw.get_file(type="Item", label="Landing_zone.jpg")
        assert f.admids == ["amdSec_2"]
        assert f.dmdids == []
        assert f.file_id() == "file-e284d015-cfb0-45dd-961d-512bf0f47cf6"
        assert f.group_id() == "Group-ab5c67fc-8f80-4e46-9f20-8d5ae29c43f2"
        assert mw.get_file(type="Directory", label="metadata") is not None
        assert mw.get_file(type="Directory", label="transfers") is not None
        assert (
            mw.get_file(
                type="Directory", label="csv-55599568-90bd-46ac-b1be-d1a538793cae"
            )
            is not None
        )
        assert (
            mw.get_file(type="Directory", label="submissionDocumentation") is not None
        )
        assert (
            mw.get_file(
                type="Directory",
                label="transfer-csv-55599568-90bd-46ac-b1be-d1a538793cae",
            )
            is not None
        )

    def test_parse_dmdsecs(self):
        """It should create FSEntry's ordered dmdsecs and mapping by type."""
        mw = metsrw.METSDocument()
        parser = etree.XMLParser(remove_blank_text=True)
        root = etree.parse("fixtures/dmdsecs_mets.xml", parser=parser)
        mw.tree = root
        mw._parse_tree()
        objects = mw.get_file(type="Directory", label="objects")
        assert len(objects.dmdsecs) == 4
        assert objects.dmdsecs[0].id_string == "dmdSec_1"
        assert objects.dmdsecs[-1].id_string == "dmdSec_4"
        assert len(objects.dmdsecs_by_mdtype["DC"]) == 1
        assert len(objects.dmdsecs_by_mdtype["OTHER_CUSTOM"]) == 1
        assert len(objects.dmdsecs_by_mdtype["OTHER_MD_A"]) == 1
        assert len(objects.dmdsecs_by_mdtype["OTHER_MD_B"]) == 1
        file = mw.get_file(type="Item", label="Landing_zone.jpg")
        assert len(file.dmdsecs) == 5
        assert file.dmdsecs[0].id_string == "dmdSec_5"
        assert file.dmdsecs[-1].id_string == "dmdSec_9"
        assert len(file.dmdsecs_by_mdtype["DC"]) == 2
        assert len(file.dmdsecs_by_mdtype["OTHER_CUSTOM"]) == 3

    def test_parse_tree_createdate_too_new(self):
        mw = metsrw.METSDocument()
        root = etree.parse("fixtures/createdate_too_new.xml")
        mw.tree = root
        with pytest.raises(metsrw.ParseError):
            mw._parse_tree()

    def test_parse_tree_no_createdate(self):
        mw = metsrw.METSDocument()
        mets_string = b"""<?xml version='1.0' encoding='UTF-8'?>
<mets xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns="http://www.loc.gov/METS/" xsi:schemaLocation="http://www.loc.gov/METS/ http://www.loc.gov/standards/mets/version18/mets.xsd">
  <metsHdr/><structMap TYPE="physical"></structMap>
</mets>
"""
        root = etree.fromstring(mets_string)
        mw.tree = root
        mw._parse_tree()
        assert mw.createdate is None

    def test_parse_no_groupid(self):
        """It should handle files with no GROUPID."""
        mw = metsrw.METSDocument().fromfile("fixtures/mets_without_groupid_in_file.xml")
        assert mw.get_file(file_uuid="db653873-d0ab-4bc1-9edb-2b6d2d84ab5a") is not None

    def test_write(self):
        mw = metsrw.METSDocument()
        # mock serialize
        parser = etree.XMLParser(remove_blank_text=True)
        root = etree.parse("fixtures/complete_mets.xml", parser=parser).getroot()
        mw.serialize = lambda fully_qualified=True: root
        mw.write("test_write.xml", pretty_print=True, encoding="ASCII")
        assert filecmp.cmp(
            "fixtures/complete_mets.xml", "test_write.xml", shallow=False
        )
        os.remove("test_write.xml")

    def test_write_encoding(self):
        """Test serialisation encodings using the ``write`` method."""
        mw = metsrw.METSDocument()
        mw.append_file(metsrw.FSEntry("path", file_uuid=str(uuid.uuid4())))
        fd, mets_path = tempfile.mkstemp()

        try:
            mw.write(mets_path)
            with open(mets_path, "rb") as file_obj:
                tree = etree.parse(file_obj)
            assert tree.docinfo.encoding == "UTF-8"
        finally:
            os.close(fd)
            os.remove(mets_path)

    def test_tostring_encoding(self):
        """Test serialisation encodings using the ``tostring`` method."""
        mw = metsrw.METSDocument()
        mw.append_file(metsrw.FSEntry("path", file_uuid=str(uuid.uuid4())))

        xml = mw.tostring()
        assert isinstance(xml, bytes)
        tree = etree.parse(io.BytesIO(xml))
        assert tree.docinfo.encoding == "UTF-8"

        xml = mw.tostring(encoding="ASCII")
        assert isinstance(xml, bytes)
        tree = etree.parse(io.BytesIO(xml))
        assert tree.docinfo.encoding == "ASCII"

        xml = mw.tostring(encoding="unicode")
        assert isinstance(xml, str)

    def test_mets_root(self):
        mw = metsrw.METSDocument()
        root = mw._document_root()
        location = (
            "http://www.loc.gov/METS/ "
            + "http://www.loc.gov/standards/mets/version1121/mets.xsd"
        )
        assert root.tag == "{http://www.loc.gov/METS/}mets"
        assert root.attrib[metsrw.lxmlns("xsi") + "schemaLocation"] == location
        nsmap = {
            "mets": "http://www.loc.gov/METS/",
            "xsi": "http://www.w3.org/2001/XMLSchema-instance",
            "xlink": "http://www.w3.org/1999/xlink",
        }
        assert root.nsmap == nsmap

    def test_mets_root_attributes(self):
        mw = metsrw.METSDocument().fromfile("fixtures/mets_without_groupid_in_file.xml")
        assert mw.objid == "44db0a40-4c76-45b9-83f1-a8adba434b43"
        assert (
            mw._document_root().attrib["OBJID"]
            == "44db0a40-4c76-45b9-83f1-a8adba434b43"
        )

    def test_mets_header(self):
        mw = metsrw.METSDocument()
        date = "2014-07-16T22:52:02.480108"
        header = mw._mets_header(date)
        assert header.tag == "{http://www.loc.gov/METS/}metsHdr"
        assert header.attrib["CREATEDATE"] == date

    def test_mets_header_lastmoddate(self):
        mw = metsrw.METSDocument()
        date = "2014-07-16T22:52:02.480108"
        new_date = "3014-07-16T22:52:02.480108"
        mw.createdate = date
        header = mw._mets_header(new_date)
        assert header.tag == "{http://www.loc.gov/METS/}metsHdr"
        assert header.attrib["CREATEDATE"] == date
        assert header.attrib["LASTMODDATE"] == new_date
        assert header.attrib["CREATEDATE"] < header.attrib["LASTMODDATE"]

    def test_mets_header_with_agent(self):
        mets = metsrw.METSDocument()
        agent = metsrw.Agent(
            role="CREATOR",
            type="SOFTWARE",
            name="39461beb-22eb-4942-88af-848cfc3462b2",
            notes=["Archivematica dashboard UUID"],
        )
        mets.agents.append(agent)

        header_element = mets._mets_header("2014-07-16T22:52:02.480108")
        agent_element = header_element.find("{http://www.loc.gov/METS/}agent")

        assert agent_element.get("ROLE") == agent.role
        assert agent_element.get("TYPE") == "OTHER"
        assert agent_element.get("OTHERTYPE") == agent.type
        assert agent_element.find("{http://www.loc.gov/METS/}name").text == agent.name
        assert (
            agent_element.find("{http://www.loc.gov/METS/}note").text == agent.notes[0]
        )

    def test_parse_header_with_agent(self):
        mets = metsrw.METSDocument.fromstring(
            b"""<?xml version='1.0' encoding='UTF-8'?>
<mets xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns="http://www.loc.gov/METS/" xsi:schemaLocation="http://www.loc.gov/METS/ http://www.loc.gov/standards/mets/version18/mets.xsd">
    <metsHdr CREATEDATE="2015-12-16T22:38:48">
        <agent OTHERTYPE="SOFTWARE" ROLE="CREATOR" TYPE="OTHER">
            <name>39461beb-22eb-4942-88af-848cfc3462b2</name>
            <note>Archivematica dashboard UUID</note>
        </agent>
    </metsHdr>
    <structMap ID="structMap_1" LABEL="Archivematica default" TYPE="physical"/>
</mets>"""
        )

        assert len(mets.agents) == 1
        assert mets.agents[0].role == "CREATOR"
        assert mets.agents[0].type == "SOFTWARE"
        assert mets.agents[0].name == "39461beb-22eb-4942-88af-848cfc3462b2"
        assert mets.agents[0].notes[0] == "Archivematica dashboard UUID"

    def test_mets_header_with_alt_record_id(self):
        mets = metsrw.METSDocument()
        alt_record_id = metsrw.AltRecordID(
            "39461beb-22eb-4942-88af-848cfc3462b2", type="Accession ID"
        )
        mets.alternate_ids.append(alt_record_id)

        header_element = mets._mets_header("2014-07-16T22:52:02.480108")
        alt_record_id_element = header_element.find(
            "{http://www.loc.gov/METS/}altRecordID"
        )

        assert alt_record_id_element.get("TYPE") == alt_record_id.type
        assert alt_record_id_element.text == alt_record_id.text

    def test_parse_header_with_alt_record_id(self):
        mets = metsrw.METSDocument.fromstring(
            b"""<?xml version='1.0' encoding='ASCII'?>
<mets xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns="http://www.loc.gov/METS/" xsi:schemaLocation="http://www.loc.gov/METS/ http://www.loc.gov/standards/mets/version18/mets.xsd">
    <metsHdr CREATEDATE="2015-12-16T22:38:48">
        <altRecordID TYPE="Accession Id">39461beb-22eb-4942-88af-848cfc3462b2</altRecordID>
    </metsHdr>
    <structMap ID="structMap_1" LABEL="Archivematica default" TYPE="physical"/>
</mets>"""
        )

        assert len(mets.alternate_ids) == 1
        assert mets.alternate_ids[0].type == "Accession Id"
        assert mets.alternate_ids[0].text == "39461beb-22eb-4942-88af-848cfc3462b2"

    def test_fromfile_invalid_xlink_href(self):
        """Test that ``fromfile`` raises ``ParseError`` if an xlink:href value
        in the source METS contains an unparseable URL.
        """
        with pytest.raises(metsrw.exceptions.ParseError, match="is not a valid URL."):
            metsrw.METSDocument.fromfile("fixtures/mets_invalid_xlink_hrefs.xml")

    def test_analyze_fptr(self):
        parser = etree.XMLParser(remove_blank_text=True)
        tree = etree.parse("fixtures/mets_dir_with_fptrs.xml", parser=parser)
        mw = metsrw.METSDocument()

        # Test that exception is raised when fileSec cannot be found.
        fptr_elem = etree.fromstring('<fptr FILEID="12345"/>')
        with pytest.raises(
            metsrw.exceptions.ParseError,
            match="12345 exists in structMap but not fileSec",
        ):
            metsrw.METSDocument._analyze_fptr(fptr_elem, tree, "directory")

        # Test that exception is raised when the path cannot be decoded.
        fptr_elem = etree.fromstring(
            '<fptr FILEID="AM68.csv-fc0e52ca-a688-41c0-a10b-c1d36e21e804"/>'
        )
        with mock.patch("metsrw.utils.urldecode") as urldecode:
            urldecode.side_effect = ValueError()
            with pytest.raises(
                metsrw.exceptions.ParseError, match="is not a valid URL"
            ):
                metsrw.METSDocument._analyze_fptr(fptr_elem, tree, "directory")

        # Test the integrity of the ``FPtr`` object returned.
        fptr = mw._analyze_fptr(fptr_elem, tree, "directory")
        assert fptr == metsrw.mets.FPtr(
            fileid="AM68.csv-fc0e52ca-a688-41c0-a10b-c1d36e21e804",
            file_uuid="fc0e52ca-a688-41c0-a10b-c1d36e21e804",
            derived_from=None,
            use="original",
            path="objects/AM68.csv",
            amdids="amdSec_3",
            dmdids=None,
            checksum=None,
            checksumtype=None,
            transform_files=[],
        )

    def test_analyze_fptr_with_dmdsecs_in_filesec(self):
        parser = etree.XMLParser(remove_blank_text=True)
        tree = etree.parse("fixtures/mets_with_dmdsecs_in_filesec.xml", parser=parser)
        fptr_elem = tree.find(
            '/mets:structMap[@TYPE="physical"]//mets:div[@LABEL="bitstream_8266.pdf"]/mets:fptr',
            namespaces=metsrw.utils.NAMESPACES,
        )

        # Test the integrity of the ``FPtr`` object returned.
        mw = metsrw.METSDocument()
        fptr = mw._analyze_fptr(fptr_elem, tree, "Item")
        assert fptr == metsrw.mets.FPtr(
            fileid="file-33f5f35a-8bde-4b94-b7cd-3d2c8b8f7a23",
            file_uuid="33f5f35a-8bde-4b94-b7cd-3d2c8b8f7a23",
            derived_from=None,
            use="original",
            path="objects/ITEM_2429-2700.zip-2023-07-07T23_05_58.201656_00_00/bitstream_8266.pdf",
            amdids="amdSec_3",
            dmdids="dmdSec_3 dmdSec_4",
            checksum=None,
            checksumtype=None,
            transform_files=[],
        )

    def test_analyze_fptr_from_aip(self):
        parser = etree.XMLParser(remove_blank_text=True)
        tree = etree.parse("fixtures/production-pointer-file.xml", parser=parser)
        mw = metsrw.METSDocument()

        fptr_elem = tree.find("//mets:fptr[1]", namespaces=metsrw.utils.NAMESPACES)
        fptr = mw._analyze_fptr(fptr_elem, tree, "Archival Information Package")
        assert fptr.file_uuid == "7327b00f-d83a-4ae8-bb89-84fce994e827"
        assert fptr.use == "Archival Information Package"
        assert fptr.transform_files == [
            {"ALGORITHM": "bzip2", "ORDER": "1", "TYPE": "decompression"}
        ]

    def test_analyze_fptr_sets_uuid_from_aip_with_file_id_prefix(self):
        """
        Test that AIP FILEIDs with a leading `file-` are parsed properly.
        """
        tree = etree.fromstring(
            b"""<?xml version='1.0' encoding='utf-8'?>
<mets:mets xmlns:mets="http://www.loc.gov/METS/" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.loc.gov/METS/ http://www.loc.gov/standards/mets/version1121/mets.xsd">
  <mets:fileSec>
    <mets:fileGrp USE="Archival Information Package">
      <mets:file ID="file-example-1-9b9f129c-8062-471b-a009-9ee0ad655f08">
        <mets:FLocat OTHERLOCTYPE="SYSTEM" LOCTYPE="OTHER" xlink:href="/tmp/example-1-9b9f129c-8062-471b-a009-9ee0ad655f08.7z"/>
      </mets:file>
    </mets:fileGrp>
  </mets:fileSec>
  <mets:structMap TYPE="physical">
    <mets:div ADMID="amdSec_1" TYPE="Archival Information Package">
      <mets:fptr FILEID="file-example-1-9b9f129c-8062-471b-a009-9ee0ad655f08"/>
    </mets:div>
  </mets:structMap>
</mets:mets>
        """
        )
        fptr_elem = tree.find(".//mets:fptr[1]", namespaces=metsrw.utils.NAMESPACES)
        fptr = metsrw.METSDocument()._analyze_fptr(
            fptr_elem, tree, "Archival Information Package"
        )

        assert fptr.file_uuid == "9b9f129c-8062-471b-a009-9ee0ad655f08"

    def test_analyze_fptr_sets_uuid_from_aic_with_file_id_prefix(self):
        """
        Test that AIC FILEIDs with a leading `file-` are parsed properly.
        """
        tree = etree.fromstring(
            b"""<?xml version='1.0' encoding='utf-8'?>
<mets:mets xmlns:mets="http://www.loc.gov/METS/" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.loc.gov/METS/ http://www.loc.gov/standards/mets/version1121/mets.xsd">
  <mets:fileSec>
    <mets:fileGrp USE="Archival Information Package">
      <mets:file ID="file-aictest_replica_aic-258f73d1-08fe-4ade-9770-1d8812cde545" GROUPID="Group-258f73d1-08fe-4ade-9770-1d8812cde545" ADMID="amdSec_10">
        <mets:FLocat xlink:href="/var/archivematica/sharedDirectory/www/AIPsStore/258f/73d1/08fe/4ade/9770/1d88/12cd/e545/aictest_replica_aic-258f73d1-08fe-4ade-9770-1d8812cde545.7z" LOCTYPE="OTHER" OTHERLOCTYPE="SYSTEM"/>
        <mets:transformFile TRANSFORMTYPE="decompression" TRANSFORMORDER="1" TRANSFORMALGORITHM="bzip2"/>
      </mets:file>
    </mets:fileGrp>
  </mets:fileSec>
  <mets:structMap ID="structMap_1" LABEL="Archivematica default" TYPE="physical">
    <mets:div TYPE="Archival Information Collection" LABEL="aictest_replica_aic-258f73d1-08fe-4ade-9770-1d8812cde545.7z">
      <mets:fptr FILEID="file-aictest_replica_aic-258f73d1-08fe-4ade-9770-1d8812cde545"/>
    </mets:div>
  </mets:structMap>
  <mets:structMap ID="structMap_2" LABEL="Normative Directory Structure" TYPE="logical">
    <mets:div TYPE="Archival Information Collection" LABEL="aictest_replica_aic-258f73d1-08fe-4ade-9770-1d8812cde545.7z"/>
  </mets:structMap>
</mets:mets>
        """
        )
        fptr_elem = tree.find(".//mets:fptr[1]", namespaces=metsrw.utils.NAMESPACES)
        fptr = metsrw.METSDocument()._analyze_fptr(
            fptr_elem, tree, "Archival Information Collection"
        )

        assert fptr.file_uuid == "258f73d1-08fe-4ade-9770-1d8812cde545"

    def test_duplicate_ids(self):
        """
        We don't want duplicate ids to be generated, but if specified, they shouldn't break
        serialization.
        """
        document = metsrw.METSDocument()
        fsentry1 = metsrw.FSEntry("file[1].txt", file_uuid=str(uuid.uuid4()))
        fsentry1.add_premis_object("<premis>object</premis>")
        fsentry1.amdsecs[0].id_string = "id_1"
        fsentry2 = metsrw.FSEntry("file[2].txt", file_uuid=str(uuid.uuid4()))
        fsentry2.add_premis_object("<premis>object</premis>")
        fsentry2.amdsecs[0].id_string = "id_1"

        document.append_file(fsentry1)
        document.append_file(fsentry2)

        reloaded_document = metsrw.METSDocument.fromtree(document.serialize())
        # Third time's the charm - previously this failed
        reloaded_document = metsrw.METSDocument.fromtree(reloaded_document.serialize())
        amdsecs = reloaded_document.tree.findall("{http://www.loc.gov/METS/}amdSec")

        assert len(amdsecs) == 2
        assert amdsecs[0].get("ID") == "id_1"
        assert amdsecs[1].get("ID") == "id_1"

    def test_directory_amd_section_read(self):
        """It should populate directory AMD sections."""
        mw = metsrw.METSDocument.fromfile("fixtures/mets_directory_amd.xml")
        objects_dir = next(_ for _ in mw.all_files() if _.label == "objects")
        assert len(objects_dir.admids) == 1


class TestWholeMETS(TestCase):
    """Test integration between classes."""

    def test_files(self):
        # Test collects several children deep
        f3 = metsrw.FSEntry("level3.txt", file_uuid=str(uuid.uuid4()))
        d2 = metsrw.FSEntry("dir2", type="Directory", children=[f3])
        f2 = metsrw.FSEntry("level2.txt", file_uuid=str(uuid.uuid4()))
        d1 = metsrw.FSEntry("dir1", type="Directory", children=[d2, f2])
        f1 = metsrw.FSEntry("level1.txt", file_uuid=str(uuid.uuid4()))
        d = metsrw.FSEntry("root", type="Directory", children=[d1, f1])
        mw = metsrw.METSDocument()
        mw.append_file(d)
        files = mw.all_files()
        assert files
        assert len(files) == 6
        assert d in files
        assert f1 in files
        assert d1 in files
        assert f2 in files
        assert d2 in files
        assert f3 in files
        f4_uuid = str(uuid.uuid4())
        f4 = metsrw.FSEntry("file4.txt", file_uuid=f4_uuid)
        mw.append_file(f4)
        files = mw.all_files()
        assert len(files) == 7
        assert f4 in files

    def test_add_file_to_child(self):
        # Test collects several children deep
        f2 = metsrw.FSEntry("level2.txt", file_uuid=str(uuid.uuid4()))
        d1 = metsrw.FSEntry("dir1", type="Directory", children=[f2])
        f1 = metsrw.FSEntry("level1.txt", file_uuid=str(uuid.uuid4()))
        d = metsrw.FSEntry("root", type="Directory", children=[d1, f1])
        mw = metsrw.METSDocument()
        mw.append_file(d)
        files = mw.all_files()
        assert files
        assert len(files) == 4
        assert d in files
        assert f1 in files
        assert d1 in files
        assert f2 in files

        f3 = metsrw.FSEntry("level3.txt", file_uuid=str(uuid.uuid4()))
        d1.add_child(f3)
        files = mw.all_files()
        assert len(files) == 5
        assert f3 in files

    def test_get_file(self):
        # Setup
        f3_uuid = str(uuid.uuid4())
        f3 = metsrw.FSEntry("dir1/dir2/level3.txt", file_uuid=f3_uuid)
        d2 = metsrw.FSEntry("dir1/dir2", type="Directory", children=[f3])
        f2_uuid = str(uuid.uuid4())
        f2 = metsrw.FSEntry("dir1/level2.txt", file_uuid=f2_uuid)
        d1 = metsrw.FSEntry("dir1", type="Directory", children=[d2, f2])
        f1_uuid = str(uuid.uuid4())
        f1 = metsrw.FSEntry("level1.txt", file_uuid=f1_uuid)
        d = metsrw.FSEntry("root", type="Directory", children=[d1, f1])
        mw = metsrw.METSDocument()
        mw.append_file(d)
        # Test
        # By UUID
        assert mw.get_file(file_uuid=f3_uuid) == f3
        assert mw.get_file(file_uuid=f2_uuid) == f2
        assert mw.get_file(file_uuid=f1_uuid) == f1
        assert mw.get_file(file_uuid="does not exist") is None
        # By path
        assert mw.get_file(path="dir1/dir2/level3.txt") == f3
        assert mw.get_file(path="dir1/dir2") == d2
        assert mw.get_file(path="dir1/level2.txt") == f2
        assert mw.get_file(path="dir1") == d1
        assert mw.get_file(path="level1.txt") == f1
        assert mw.get_file(path="does not exist") is None
        # By label
        assert mw.get_file(label="level3.txt") == f3
        assert mw.get_file(label="dir2") == d2
        assert mw.get_file(label="level2.txt") == f2
        assert mw.get_file(label="dir1") == d1
        assert mw.get_file(label="level1.txt") == f1
        assert mw.get_file(label="does not exist") is None
        # By multiple
        assert mw.get_file(label="level3.txt", path="dir1/dir2/level3.txt") == f3
        assert mw.get_file(label="dir2", type="Directory") == d2
        assert mw.get_file(label="level2.txt", type="Item") == f2
        assert mw.get_file(file_uuid=None, type="Item") is None
        # Updates list
        f4_uuid = str(uuid.uuid4())
        f4 = metsrw.FSEntry("file4.txt", file_uuid=f4_uuid)
        mw.append_file(f4)
        assert mw.get_file(file_uuid=f4_uuid) == f4
        assert mw.get_file(path="file4.txt") == f4

    def test_remove_file(self):
        """It should"""
        # Setup
        f3_uuid = str(uuid.uuid4())
        f3 = metsrw.FSEntry("dir1/dir2/level3.txt", file_uuid=f3_uuid)
        d2 = metsrw.FSEntry("dir1/dir2", type="Directory", children=[f3])
        f2_uuid = str(uuid.uuid4())
        f2 = metsrw.FSEntry("dir1/level2.txt", file_uuid=f2_uuid)
        d1 = metsrw.FSEntry("dir1", type="Directory", children=[d2, f2])
        f1_uuid = str(uuid.uuid4())
        f1 = metsrw.FSEntry("level1.txt", file_uuid=f1_uuid)
        d = metsrw.FSEntry("root", type="Directory", children=[d1, f1])
        mw = metsrw.METSDocument()
        mw.append_file(d)
        assert len(mw.all_files()) == 6
        # Test remove file
        mw.remove_entry(f3)
        assert len(mw.all_files()) == 5
        assert mw.get_file(file_uuid=f3_uuid) is None
        assert f3 not in d2.children
        assert f3 not in mw.all_files()
        # Test remove dir
        mw.remove_entry(d1)
        assert len(mw.all_files()) == 2
        assert mw.get_file(path="dir1") is None
        assert d1 not in d.children
        assert d1 not in mw.all_files()
        assert f2 not in mw.all_files()
        assert d2 not in mw.all_files()
        assert f1 in d.children
        # Test remove root element
        mw.remove_entry(d)
        assert len(mw.all_files()) == 0

    def test_collect_mdsec_elements(self):
        f1 = metsrw.FSEntry("file1.txt", file_uuid=str(uuid.uuid4()))
        f1.amdsecs.append(metsrw.AMDSec())
        f1.dmdsecs.append(metsrw.SubSection("dmdSec", None))
        f2 = metsrw.FSEntry("file2.txt", file_uuid=str(uuid.uuid4()))
        f2.dmdsecs.append(metsrw.SubSection("dmdSec", None))
        mw = metsrw.METSDocument()
        elements = mw._collect_mdsec_elements([f1, f2])
        # Check ordering - dmdSec before amdSec
        assert isinstance(elements, list)
        assert len(elements) == 3
        assert isinstance(elements[0], metsrw.SubSection)
        assert elements[0].subsection == "dmdSec"
        assert isinstance(elements[1], metsrw.SubSection)
        assert elements[1].subsection == "dmdSec"
        assert isinstance(elements[2], metsrw.AMDSec)

    def test_filesec(self):
        o = metsrw.FSEntry("objects/file1.txt", file_uuid=str(uuid.uuid4()))
        p = metsrw.FSEntry(
            "objects/file1-preservation.txt",
            use="preservaton",
            file_uuid=str(uuid.uuid4()),
        )
        o2 = metsrw.FSEntry("objects/file2.txt", file_uuid=str(uuid.uuid4()))
        mw = metsrw.METSDocument()
        element = mw._filesec([o, p, o2])
        assert isinstance(element, etree._Element)
        assert element.tag == "{http://www.loc.gov/METS/}fileSec"
        assert len(element) == 2  # 2 groups
        assert element[0].tag == "{http://www.loc.gov/METS/}fileGrp"
        assert element[0].get("USE") == "original"
        assert element[1].tag == "{http://www.loc.gov/METS/}fileGrp"
        assert element[1].get("USE") == "preservaton"
        # TODO test file & FLocat

    def test_structmap(self):
        """
        It should create a structMap tag.
        It should have a div tag for the directory.
        It should have div tags for the children beneath the directory.
        It should not have div tags for deleted files (without label).
        """
        children = [
            metsrw.FSEntry("objects/file1.txt", file_uuid=str(uuid.uuid4())),
            metsrw.FSEntry("objects/file2.txt", file_uuid=str(uuid.uuid4())),
        ]
        parent = metsrw.FSEntry("objects", type="Directory", children=children)
        deleted_f = metsrw.FSEntry(use="deletion", file_uuid=str(uuid.uuid4()))

        writer = metsrw.METSDocument()
        writer.append_file(parent)
        writer.append_file(deleted_f)
        sm = writer._structmap()

        assert sm.tag == "{http://www.loc.gov/METS/}structMap"
        assert sm.attrib["TYPE"] == "physical"
        assert sm.attrib["ID"] == "structMap_1"
        assert sm.attrib["LABEL"] == "Archivematica default"
        assert len(sm.attrib) == 3
        assert len(sm) == 1
        parent = sm[0]
        assert parent.tag == "{http://www.loc.gov/METS/}div"
        assert parent.attrib["LABEL"] == "objects"
        assert parent.attrib["TYPE"] == "Directory"
        assert len(parent.attrib) == 2
        assert len(parent) == 2
        assert parent[0].attrib["LABEL"] == "file1.txt"
        assert parent[0].attrib["TYPE"] == "Item"
        assert len(parent[0].attrib) == 2
        assert parent[0].find("{http://www.loc.gov/METS/}fptr") is not None
        assert parent[1].attrib["LABEL"] == "file2.txt"
        assert parent[1].attrib["TYPE"] == "Item"
        assert len(parent[1].attrib) == 2
        assert parent[1].find("{http://www.loc.gov/METS/}fptr") is not None

    def test_full_mets(self):
        mw = metsrw.METSDocument()
        file1 = metsrw.FSEntry("objects/object1.ext", file_uuid=str(uuid.uuid4()))
        file2 = metsrw.FSEntry("objects/object2.ext", file_uuid=str(uuid.uuid4()))
        file1p = metsrw.FSEntry(
            "objects/object1-preservation.ext",
            use="preservation",
            file_uuid=str(uuid.uuid4()),
            derived_from=file1,
        )
        file2p = metsrw.FSEntry(
            "objects/object2-preservation.ext",
            use="preservation",
            file_uuid=str(uuid.uuid4()),
            derived_from=file2,
        )
        children = [file1, file2, file1p, file2p]
        objects = metsrw.FSEntry("objects", type="Directory", children=children)
        children = [
            metsrw.FSEntry("transfers", type="Directory", children=[]),
            metsrw.FSEntry(
                "metadata/metadata.csv", use="metadata", file_uuid=str(uuid.uuid4())
            ),
        ]
        metadata = metsrw.FSEntry("metadata", type="Directory", children=children)
        children = [
            metsrw.FSEntry(
                "submissionDocumentation/METS.xml",
                use="submissionDocumentation",
                file_uuid=str(uuid.uuid4()),
            )
        ]
        sub_doc = metsrw.FSEntry(
            "submissionDocumentation", type="Directory", children=children
        )
        children = [objects, metadata, sub_doc]
        sip = metsrw.FSEntry("sipname-uuid", type="Directory", children=children)
        sip.add_dublin_core("<dublincore>sip metadata</dublincore>")
        file1.add_premis_object("<premis>object</premis>")
        file1.add_premis_event("<premis>event</premis>")
        file1.add_premis_agent("<premis>agent</premis>")
        rights = file1.add_premis_rights("<premis>rights</premis>")
        rights.replace_with(file1.add_premis_rights("<premis>newer rights</premis>"))
        dc = file1.add_dublin_core("<dublincore>metadata</dublincore>")
        dc.replace_with(
            file1.add_dublin_core("<dublincore>newer metadata</dublincore>")
        )

        mw.append_file(sip)
        mw.write("full_metsrw.xml", fully_qualified=True, pretty_print=True)
        os.remove("full_metsrw.xml")

    def test_pointer_file(self):
        """Test the creation of pointer files."""

        # Mocks of the AIP, its compression event, and other details.
        aip_uuid = str(uuid.uuid4())
        aip = {
            "current_path": f"/path/to/myaip-{aip_uuid}.7z",
            "uuid": aip_uuid,
            "package_type": "Archival Information Package",
            "checksum_algorithm": "sha256",
            "checksum": "78e4509313928d2964fe877a6a82f1ba728c171eedf696e3f5b0aed61ec547f6",
            "size": "11854",
            "extension": ".7z",
            "archive_tool": "7-Zip",
            "archive_tool_version": "9.20",
            "transform_files": [
                {"algorithm": "bzip2", "order": "2", "type": "decompression"},
                {"algorithm": "gpg", "order": "1", "type": "decryption"},
            ],
        }
        compression_event = {
            "uuid": str(uuid.uuid4()),
            "detail": (
                "program=7z; version=p7zip Version 9.20"
                " (locale=en_US.UTF-8,Utf16=on,HugeFiles=on,2 CPUs)"
            ),
            "outcome": "",
            # This should be the output from 7-zip or other...
            "outcome_detail_note": "",
            "agents": [
                {
                    "name": "Archivematica",
                    "type": "software",
                    "identifier_type": "preservation system",
                    "identifier_value": "Archivematica-1.6.1",
                },
                {
                    "name": "test",
                    "type": "organization",
                    "identifier_type": "repository code",
                    "identifier_value": "test",
                },
            ],
        }
        now = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        pronom_conversion = {
            ".7z": {"puid": "fmt/484", "name": "7Zip format"},
            ".bz2": {"puid": "x-fmt/268", "name": "BZIP2 Compressed Archive"},
        }

        # Create the METS using metsrw
        mw = metsrw.METSDocument()

        # TODO: metsrw will prefix "file-" to the AIP UUID when creating <mets:fptr
        # FILEID> and <mets:file ID> attr vals. However, we want "file-" to be
        # replaced by the AIP's (i.e., the transfer's) name.
        aip_fs_entry = metsrw.FSEntry(
            path=aip["current_path"],
            file_uuid=aip["uuid"],
            use=aip["package_type"],
            type=aip["package_type"],
            transform_files=aip["transform_files"],
        )

        premis_schema_location = (
            "info:lc/xmlns/premis-v2"
            " http://www.loc.gov/standards/premis/v2/premis-v2-2.xsd"
        )
        nsmap = {
            "mets": metsrw.NAMESPACES["mets"],
            "xsi": metsrw.NAMESPACES["xsi"],
            "xlink": metsrw.NAMESPACES["xlink"],
        }
        E_P = ElementMaker(
            namespace=metsrw.NAMESPACES["premis"],
            nsmap={"premis": metsrw.NAMESPACES["premis"]},
        )

        # Create the AIP's PREMIS:OBJECT using raw lxml
        aip_premis_object = E_P.object(
            E_P.objectIdentifier(
                E_P.objectIdentifierType("UUID"), E_P.objectIdentifierValue(aip["uuid"])
            ),
            E_P.objectCharacteristics(
                E_P.compositionLevel("1"),
                E_P.fixity(
                    E_P.messageDigestAlgorithm(aip["checksum_algorithm"]),
                    E_P.messageDigest(aip["checksum"]),
                ),
                E_P.size(str(aip["size"])),
                E_P.format(
                    E_P.formatDesignation(
                        E_P.formatName(pronom_conversion[aip["extension"]]["name"]),
                        E_P.formatVersion(),
                    ),
                    E_P.formatRegistry(
                        E_P.formatRegistryName("PRONOM"),
                        E_P.formatRegistryKey(
                            pronom_conversion[aip["extension"]]["puid"]
                        ),
                    ),
                ),
                E_P.creatingApplication(
                    E_P.creatingApplicationName(aip["archive_tool"]),
                    E_P.creatingApplicationVersion(aip["archive_tool_version"]),
                    E_P.dateCreatedByApplication(now),
                ),
            ),
            version="2.2",
        )
        aip_premis_object.attrib["{" + nsmap["xsi"] + "}type"] = "premis:file"
        aip_premis_object.attrib["{" + nsmap["xsi"] + "}schemaLocation"] = (
            premis_schema_location
        )
        aip_fs_entry.add_premis_object(aip_premis_object)

        # Create the AIP's PREMIS:EVENT for the compression using raw lxml
        aip_premis_compression_event = E_P.event(
            E_P.eventIdentifier(
                E_P.eventIdentifierType("UUID"),
                E_P.eventIdentifierValue(compression_event["uuid"]),
            ),
            E_P.eventType("compression"),
            E_P.eventDateTime(now),
            E_P.eventDetail(compression_event["detail"]),
            E_P.eventOutcomeInformation(
                E_P.eventOutcome(compression_event["outcome"]),
                E_P.eventOutcomeDetail(
                    E_P.eventOutcomeDetailNote(compression_event["outcome_detail_note"])
                ),
            ),
            *[
                E_P.linkingAgentIdentifier(
                    E_P.linkingAgentIdentifierType(ag["identifier_type"]),
                    E_P.linkingAgentIdentifierValue(ag["identifier_value"]),
                )
                for ag in compression_event["agents"]
            ],
            version="2.2",
        )
        aip_premis_compression_event.attrib["{" + nsmap["xsi"] + "}schemaLocation"] = (
            premis_schema_location
        )
        aip_fs_entry.add_premis_event(aip_premis_compression_event)

        # Create the AIP's PREMIS:AGENTs using raw lxml
        for agent in compression_event["agents"]:
            agent_el = E_P.agent(
                E_P.agentIdentifier(
                    E_P.agentIdentifierType(agent["identifier_type"]),
                    E_P.agentIdentifierValue(agent["identifier_value"]),
                ),
                E_P.agentName(agent["name"]),
                E_P.agentType(agent["type"]),
            )
            agent_el.attrib["{" + nsmap["xsi"] + "}schemaLocation"] = (
                premis_schema_location
            )
            aip_fs_entry.add_premis_agent(agent_el)

        mw.append_file(aip_fs_entry)
        self.assert_pointer_valid(mw.serialize())

    def test_production_mets_file(self):
        mets_path = "fixtures/production-aip-mets-file.xml"
        mets_doc = etree.parse(mets_path)
        self.assert_mets_valid(mets_doc)

    def test_production_pointer_file(self):
        mets_path = "fixtures/production-pointer-file.xml"
        mets_doc = etree.parse(mets_path)
        self.assert_pointer_valid(mets_doc)

    def test_parse_production_pointer_file(self):
        """Test that we can use ``get_file`` to get the FSEntry instance
        representing the AIP using the AIP's UUID. This is made challenging by
        the fact that in pointer files the FILEID attribute's value is NOT
        prefixed by "file-".
        """
        mets_path = "fixtures/production-pointer-file.xml"
        mw = metsrw.METSDocument.fromfile(mets_path)
        aip_uuid = "7327b00f-d83a-4ae8-bb89-84fce994e827"
        assert mw.get_file(file_uuid=aip_uuid)

    def test_parse_dir_with_fptrs(self):
        mets_path = "fixtures/mets_dir_with_fptrs.xml"
        mw = metsrw.METSDocument.fromfile(mets_path)
        assert len(mw.all_files()) == 5
        assert mw.get_file(type="Directory", label="objects")
        for item in (
            ["3a6a182a-40a0-4c2b-9752-fc7e91ac1edf", "objects/V00154.MPG"],
            ["431913ba-4379-4373-8798-cc5f2b9dd769", "objects/V00158.MPG"],
            ["fc0e52ca-a688-41c0-a10b-c1d36e21e804", "objects/AM68.csv"],
        ):
            assert mw.get_file(type="Item", file_uuid=item[0], path=item[1])

    def test_mets_file_with_rights(self):
        mets_path = "fixtures/transfer_mets.xml"
        mw = metsrw.METSDocument.fromfile(mets_path)

        fsentry = mw.get_file(type="Item", label="bird.mp3")
        assert len(mw.all_files()) == 11

        rights = fsentry.get_premis_rights()
        assert len(rights) == 2
        assert not fsentry.get_premis_rights_statement("<unknown>")

        rights_statement = fsentry.get_premis_rights_statement(
            "86f63d9a-aea4-4640-807b-f0c4fa33ae88"
        )
        assert rights_statement.rights_basis == "Other"
        assert len(rights_statement.rights_granted) == 1
        assert rights_statement.act == "Act donor"

    # Helper methods

    def assert_mets_valid(self, mets_doc, schematron=metsrw.AM_SCT_PATH):
        is_valid, report = metsrw.validate(mets_doc, schematron=schematron)
        if not is_valid:
            raise AssertionError(report["report"])

    def assert_pointer_valid(self, mets_doc):
        self.assert_mets_valid(mets_doc, schematron=metsrw.AM_PNTR_SCT_PATH)

    def test_read_method_and_sequence_behaviour(self):
        mets_path = "fixtures/complete_mets.xml"

        # .read works with string, bytes, path and file descriptor.
        mets_file = open(mets_path)
        with open(mets_path) as fi:
            mets_unicode = fi.read()
            mets_bytes = mets_unicode.encode("utf8")
        mets1 = metsrw.METSDocument.read(mets_path)
        mets2 = metsrw.METSDocument.read(mets_file)
        mets3 = metsrw.METSDocument.read(mets_unicode)
        mets4 = metsrw.METSDocument.read(mets_bytes)

        # iteration
        for fse1, fse2, fse3, fse4 in zip(mets1, mets2, mets3, mets4):
            assert fse1.path == fse2.path == fse3.path == fse4.path

        # len
        assert len(mets1) == len(mets2) == len(mets3) == len(mets4)

        # indexing
        assert mets1[1].path == mets2[1].path == mets3[1].path == mets4[1].path

        # slicing
        assert (
            [fse.path for fse in mets1[:2]]
            == [fse.path for fse in mets2[:2]]
            == [fse.path for fse in mets3[:2]]
            == [fse.path for fse in mets4[:2]]
        )

    def test_files_invalid_path(self):
        """Test that if you try to set the path of a FSEntry to something that
        urllib.urlparse cannot parse and then attempt to serialize the METS,
        then you will trigger a MetsError.
        """
        f1 = metsrw.FSEntry(
            "http://foo[bar.com/hello[1].pdf", file_uuid=str(uuid.uuid4())
        )
        mw = metsrw.METSDocument()
        mw.append_file(f1)
        with pytest.raises(
            metsrw.exceptions.SerializeError, match="is not a valid URL."
        ):
            mw.serialize()

    def test_serialize_normative_structmap(self):
        """Test METS serialization with and without a normative structmap."""
        mw = metsrw.METSDocument.fromfile("fixtures/mets_empty_dirs.xml")
        xpath = (
            'mets:structMap[@LABEL="Normative Directory Structure"][@TYPE="logical"]'
        )
        tree = mw.serialize()
        assert tree.find(xpath, namespaces=metsrw.NAMESPACES) is not None
        tree = mw.serialize(normative_structmap=False)
        assert tree.find(xpath, namespaces=metsrw.NAMESPACES) is None

    def test_dspace_mets_dmdsecs(self):
        mets_path = "fixtures/mets_with_dmdsecs_in_filesec.xml"
        mw = metsrw.METSDocument.fromfile(mets_path)
        fsentry = mw.get_file(
            type="Item",
            path="objects/ITEM_2429-2700.zip-2023-07-07T23_05_58.201656_00_00/bitstream_8266.pdf",
        )

        # The original object of a DSpace transfer should contain two dmdSecs:
        assert len(fsentry.dmdsecs) == 2

        # - The first contains Xpointers to descriptive metadata in
        #   the original mets.xml files exported from DSpace.
        xpointer_dmdsec = [
            dmdsec
            for dmdsec in fsentry.dmdsecs
            if isinstance(dmdsec.contents, metsrw.MDRef)
        ][0].serialize()
        assert xpointer_dmdsec.attrib.get("STATUS") == "original"
        assert xpointer_dmdsec.attrib.get("CREATED") == "2023-07-07T23:06:15"

        xpointer = xpointer_dmdsec.find(
            "mets:mdRef",
            namespaces=metsrw.utils.NAMESPACES,
        )
        assert xpointer is not None
        assert (
            xpointer.attrib.get("LABEL")
            == "mets.xml-Group-33f5f35a-8bde-4b94-b7cd-3d2c8b8f7a23"
        )
        assert xpointer.attrib.get("MDTYPE") == "OTHER"
        assert xpointer.attrib.get("LOCTYPE") == "OTHER"
        assert xpointer.attrib.get("OTHERLOCTYPE") == "SYSTEM"
        assert xpointer.attrib.get("XPTR") == "xpointer(id('dmdSec_366 dmdSec_367'))"

        # - The second dmdSec reflects the parent-child relationship between a
        #   DSpace object and its collection, using the handles as identifiers.
        dc_dmdsec = [
            dmdsec
            for dmdsec in fsentry.dmdsecs
            if isinstance(dmdsec.contents, metsrw.MDWrap)
        ][0].serialize()
        assert dc_dmdsec.attrib.get("STATUS") == "original"
        assert dc_dmdsec.attrib.get("CREATED") == "2023-07-07T23:06:15"

        identifier = dc_dmdsec.find(
            'mets:mdWrap[@MDTYPE="DC"]/mets:xmlData/dcterms:dublincore/dc:identifier',
            namespaces=metsrw.utils.NAMESPACES,
        )
        assert identifier is not None
        assert identifier.text == "hdl:2429/2700"

        terms = dc_dmdsec.find(
            'mets:mdWrap[@MDTYPE="DC"]/mets:xmlData/dcterms:dublincore/dcterms:isPartOf',
            namespaces=metsrw.utils.NAMESPACES,
        )
        assert terms is not None
        assert terms.text == "hdl:2429/1314"

    def test_dspace_mets_amdsec(self):
        mets_path = "fixtures/mets_with_dmdsecs_in_filesec.xml"
        mw = metsrw.METSDocument.fromfile(mets_path)
        fsentry = mw.get_file(
            type="Item",
            path="objects/ITEM_2429-2700.zip-2023-07-07T23_05_58.201656_00_00/bitstream_8266.pdf",
        )

        # The original object of a DSpace transfer should contain Xpointers
        # to rights metadata in the original mets.xml files exported from
        # DSpace.
        rights_md = [
            subsection.contents
            for subsection in fsentry.amdsecs[0].subsections
            if isinstance(subsection.contents, metsrw.MDRef)
        ][0].serialize()
        assert (
            rights_md.attrib.get("LABEL")
            == "mets.xml-988d7030-3cde-43f2-ac1f-2b8cf9d5a70b"
        )
        assert rights_md.attrib.get("MDTYPE") == "OTHER"
        assert rights_md.attrib.get("OTHERMDTYPE") == "METSRIGHTS"
        assert rights_md.attrib.get("LOCTYPE") == "OTHER"
        assert rights_md.attrib.get("OTHERLOCTYPE") == "SYSTEM"
        assert (
            rights_md.attrib.get("XPTR")
            == "xpointer(id('rightsMD_371 rightsMD_374 rightsMD_384 rightsMD_393 rightsMD_401 rightsMD_409 rightsMD_417 rightsMD_425'))"
        )

    def test_dspace_filegrp_sorting_in_filesec(self):
        mets_path = "fixtures/mets_with_dmdsecs_in_filesec.xml"
        mw = metsrw.METSDocument.fromfile(mets_path)
        filesec = mw._filesec()

        assert [filegrp.attrib["USE"] for filegrp in filesec] == [
            "original",
            "submissionDocumentation",
            "preservation",
            "license",
            "text/ocr",
        ]


@pytest.mark.parametrize(
    "mets_path, expected_counts",
    [
        (
            "fixtures/complete_mets.xml",
            {
                "dmdSec": 2,
                "amdSec": 7,
                "techMD": 7,
                "rightsMD": 0,
                "digiprovMD": 54,
                "sourceMD": 0,
            },
        ),
        (
            "fixtures/mets_all_rights.xml",
            {
                "dmdSec": 0,
                "amdSec": 2,
                "techMD": 2,
                "rightsMD": 5,
                "digiprovMD": 0,
                "sourceMD": 0,
            },
        ),
        (
            "fixtures/mets_directory_amd.xml",
            {
                "dmdSec": 1,
                "amdSec": 2,
                "techMD": 0,
                "rightsMD": 0,
                "digiprovMD": 0,
                "sourceMD": 1,
            },
        ),
    ],
)
def test_get_subsections_counts(mets_path, expected_counts):
    mw = metsrw.METSDocument().fromfile(mets_path)
    assert mw.get_subsections_counts() == expected_counts


@pytest.mark.parametrize(
    "file_group_uses,expected_uses_order",
    [
        (
            [
                "unknown2",
                "original",
                "unknown1",
                "text/ocr",
            ],
            [
                "original",
                "text/ocr",
                "unknown2",
                "unknown1",
            ],
        ),
        (
            [
                "unknown2",
                "unknown1",
                "original",
                "unknown3",
            ],
            [
                "original",
                "unknown2",
                "unknown1",
                "unknown3",
            ],
        ),
    ],
)
def test_filegrp_sorting_returns_non_default_groups(
    file_group_uses, expected_uses_order
):
    file_groups = {
        use: etree.Element(metsrw.utils.lxmlns("mets") + "fileGrp", USE=use)
        for use in file_group_uses
    }

    mw = metsrw.METSDocument()
    result = mw._sort_filegrps(file_groups)

    assert [g.attrib["USE"] for g in result] == expected_uses_order


def test_document_ignores_missing_bag_metadata_amdsec_on_the_objects_directory():
    mets_path = "fixtures/mets_with_missing_bag_metadata_amdsec.xml"

    mw = metsrw.METSDocument.fromfile(mets_path)

    objects_directory = mw.get_file(
        type="Directory", label="objects", path=None, use=None
    )
    assert not objects_directory.amdsecs



"""METS reader and writer."""

import logging

from . import plugins
from .di import Dependency
from .di import FeatureBroker
from .di import feature_broker
from .di import has_class_methods
from .di import has_methods
from .di import is_class
from .di import set_feature_broker_to_default_state
from .exceptions import MetsError
from .exceptions import ParseError
from .fsentry import FSEntry
from .metadata import Agent
from .metadata import AltRecordID
from .metadata import AMDSec
from .metadata import MDRef
from .metadata import MDWrap
from .metadata import SubSection
from .mets import METSDocument
from .utils import FILE_ID_PREFIX
from .utils import GROUP_ID_PREFIX
from .utils import NAMESPACES
from .utils import SCHEMA_LOCATIONS
from .utils import generate_mdtype_key
from .utils import lxmlns
from .utils import urldecode
from .utils import urlencode
from .validate import AM_PNTR_SCT_PATH
from .validate import AM_SCT_PATH
from .validate import METS_XSD_PATH
from .validate import get_schematron
from .validate import get_xmlschema
from .validate import report_string
from .validate import schematron_validate
from .validate import sct_report_string
from .validate import validate
from .validate import xsd_error_log_string
from .validate import xsd_validate

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())
__version__ = "0.6.1"

__all__ = [
    "Agent",
    "AltRecordID",
    "AMDSec",
    "AM_PNTR_SCT_PATH",
    "AM_SCT_PATH",
    "Dependency",
    "FILE_ID_PREFIX",
    "FSEntry",
    "FeatureBroker",
    "GROUP_ID_PREFIX",
    "MDRef",
    "MDWrap",
    "METSDocument",
    "METS_XSD_PATH",
    "MetsError",
    "NAMESPACES",
    "ParseError",
    "SCHEMA_LOCATIONS",
    "SubSection",
    "__version__",
    "feature_broker",
    "generate_mdtype_key",
    "get_schematron",
    "get_xmlschema",
    "has_class_methods",
    "has_methods",
    "is_class",
    "lxmlns",
    "plugins",
    "report_string",
    "schematron_validate",
    "sct_report_string",
    "set_feature_broker_to_default_state",
    "urldecode",
    "urlencode",
    "validate",
    "xsd_error_log_string",
    "xsd_validate",
]
